"""Unit tests for djblets.extensions.hooks.TemplateHook."""

from django.template import Context, RequestContext, Template
from django.test.client import RequestFactory
from kgb import SpyAgency
from mock import Mock

from djblets.extensions.extension import Extension
from djblets.extensions.hooks import TemplateHook
from djblets.extensions.testing import ExtensionTestCaseMixin
from djblets.testing.testcases import TestCase


class MyTestExtension(Extension):
    def initialize(self):
        self.hook_no_applies_name = 'template-hook-no-applies-name'
        self.template_hook_no_applies = TemplateHook(
            self,
            self.hook_no_applies_name,
            'test_module/some_template.html',
            [])

        self.hook_with_applies_name = 'template-hook-with-applies-name'
        self.template_hook_with_applies = TemplateHook(
            self,
            self.hook_with_applies_name,
            'test_module/some_template.html',
            [
                'test-url-name',
                'url_2',
                'url_3',
            ]
        )


class TemplateHookTests(SpyAgency, ExtensionTestCaseMixin, TestCase):
    """Unit tests for djblets.extensions.hooks.TemplateHook."""

    extension_class = MyTestExtension

    def setUp(self):
        super(TemplateHookTests, self).setUp()

        self.request = Mock()
        self.request._djblets_extensions_kwargs = {}
        self.request.path_info = '/'
        self.request.resolver_match = Mock()
        self.request.resolver_match.url_name = 'root'

    def test_hook_added_to_class_by_name(self):
        """Testing TemplateHook registration"""
        self.assertIn(
            self.extension.template_hook_with_applies,
            TemplateHook._by_name[self.extension.hook_with_applies_name])

        # The TemplateHook should also be added to the Extension's collection
        # of hooks.
        self.assertIn(self.extension.template_hook_with_applies,
                      self.extension.hooks)

    def test_hook_shutdown(self):
        """Testing TemplateHook shutdown"""
        self.extension.template_hook_with_applies.disable_hook()
        self.assertNotIn(
            self.extension.template_hook_with_applies,
            TemplateHook._by_name[self.extension.hook_with_applies_name])

        # The TemplateHook should still be in the Extension's collection
        # of hooks.
        self.assertIn(self.extension.template_hook_with_applies,
                      self.extension.hooks)

    def test_applies_to_default(self):
        """Testing TemplateHook.applies_to defaults to everything"""
        self.assertTrue(
            self.extension.template_hook_no_applies.applies_to(self.request))
        self.assertTrue(
            self.extension.template_hook_no_applies.applies_to(None))

    def test_applies_to(self):
        """Testing TemplateHook.applies_to customization"""
        self.assertFalse(
            self.extension.template_hook_with_applies.applies_to(self.request))

        self.request.resolver_match.url_name = 'test-url-name'
        self.assertTrue(
            self.extension.template_hook_with_applies.applies_to(self.request))

    def test_render_to_string(self):
        """Testing TemplateHook.render_to_string"""
        hook = TemplateHook(
            self.extension,
            name='test',
            template_name='deco/box.html',
            extra_context={
                'content': 'Hello world',
            })

        request = RequestFactory().request()
        result = hook.render_to_string(request, RequestContext(request, {
            'classname': 'test',
        }))

        self.assertHTMLEqual(
            result,
            '<div class="box-container">'
            ' <div class="box test">'
            '  <div class="box-inner">'
            '   Hello world'
            '  </div>'
            ' </div>'
            '</div>')

    def test_context_doesnt_leak(self):
        """Testing TemplateHook's context won't leak state"""
        class MyTemplateHook(TemplateHook):
            def render_to_string(self, request, context):
                context['leaky'] = True

                return ''

        MyTemplateHook(self.extension, 'test')
        context = Context({})
        context['request'] = None

        t = Template(
            '{% load djblets_extensions %}'
            '{% template_hook_point "test" %}')
        t.render(context).strip()

        self.assertNotIn('leaky', context)

    def test_render_to_string_sandbox(self):
        """Testing TemplateHook sandboxing"""
        class MyTemplateHook(TemplateHook):
            def render_to_string(self, request, context):
                raise Exception('Oh noes')

        MyTemplateHook(self.extension, 'test')
        context = Context({})
        context['request'] = None

        t = Template(
            '{% load djblets_extensions %}'
            '{% template_hook_point "test" %}')
        t.render(context).strip()

        # Didn't crash. We're good.

    def test_applies_to_sandbox(self):
        """Testing TemplateHook for applies_to"""
        class MyTemplateHook(TemplateHook):
            def applies_to(self, request):
                raise Exception

        hook = MyTemplateHook(extension=self.extension, name='test')
        context = Context({})
        context['request'] = self.request

        self.spy_on(hook.applies_to)

        t = Template(
            '{% load djblets_extensions %}'
            '{% template_hook_point "test" %}')
        string = t.render(context).strip()

        self.assertEqual(string, '')

        self.assertTrue(hook.applies_to.called)



"""App configuration for djblets.privacy."""

from django.apps import AppConfig


class PrivacyAppConfig(AppConfig):
    """Default app configuration for djblets.privacy."""

    name = 'djblets.privacy'
    label = 'djblets_privacy'



"""Unit tests for djblets.forms.fields.ListEditField."""

from djblets.forms.fields import ListEditField
from djblets.testing.testcases import TestCase


class ListEditFieldTests(TestCase):
    """Unit tests for djblets.forms.fields.ListEditField."""

    def test_prepare_value(self):
        """Testing ListEditField.prepare_value"""
        field = ListEditField()

        self.assertEqual(
            field.prepare_value(' foo,  bar , baz '),
            ['foo', 'bar', 'baz'])

    def test_prepare_value_with_custom_separator(self):
        """Testing ListEditField.prepare_value with a custom separator"""
        field = ListEditField(sep=';')

        self.assertEqual(
            field.prepare_value(' foo;  bar ; baz '),
            ['foo', 'bar', 'baz'])

    def test_prepare_value_with_null_value(self):
        """Testing ListEditField.prepare_value with a null value"""
        field = ListEditField()

        self.assertEqual(
            field.prepare_value(None),
            [''])

    def test_prepare_value_with_list_value(self):
        """Testing ListEditField.prepare_value with a list"""
        field = ListEditField()

        self.assertEqual(
            field.prepare_value(['foo', 'bar', 'baz']),
            ['foo', 'bar', 'baz'])

    def test_to_python(self):
        """Testing ListEditField.to_python"""
        field = ListEditField()

        self.assertEqual(
            field.to_python(['foo', 'bar', 'baz']),
            'foo,bar,baz')

    def test_to_python_with_custom_separator(self):
        """Testing ListEditField.to_python with a custom separator"""
        field = ListEditField(sep=';')

        self.assertEqual(
            field.to_python(['foo', 'bar', 'baz']),
            'foo;bar;baz')



"""An avatar service for setting absolute URLs for avatars."""

from django.utils.translation import gettext_lazy as _

from djblets.avatars.services.base import AvatarService


class URLAvatarService(AvatarService):
    """An avatar service for settings absolute URLs for avatars.

    This avatar service is not available to users and is meant to be used by
    bots from extensions to provide their own avatars.

    Automation users can be configured to use this service with the following
    code:

    .. code-block:: python

       from djblets.avatars.services import URLAvatarService


       avatar_services = get_avatar_service_registry()
       service = avatar_services.get_avatar_service(
           URLAvatarService.avatar_service_id)

       service.setup(
           user,
           {
               '1x': 'http://example.com/static/avatar.png',
               '2x': 'http://example.com/static/avatar@2x.png',
           })
    """

    avatar_service_id = 'url'
    name = _('URL Avatar Service')
    hidden = True

    def setup(self, user, urls):
        """Set up this avatar service for the given user.

        The user will be configured to use this service as their avatar service
        and the given URLs will be used for their avatars.

        Args:
            user (django.contrib.auth.models.User):
                The user to set the URLs for.

            urls (dict):
                A dictionary mapping resolutions ('1x', '2x', etc.) to absolute
                URLs.
        """
        settings_manager = self._settings_manager_class(user)
        settings_manager.avatar_service_id = self.avatar_service_id
        settings_manager.configuration_for(self.avatar_service_id).update(urls)
        settings_manager.save()

    def get_avatar_urls_uncached(self, user, size):
        """Return the avatar URLs for the requested user.

        Args:
            user (django.contrib.auth.models.User):
                The user whose avatar URLs are to be fetched.

            size (int):
                The size (in pixels) the avatar is to be rendered at.

        Returns:
            dict:
            A dictionary mapping resolutions ('1x', '2x', etc.) to the user's
            avatar URLs.
        """
        return (
            self._settings_manager_class(user)
            .configuration_for(self.avatar_service_id)
        )

    def get_etag_data(self, user):
        """Return the ETag data for the user's avatar.

        Args:
            user (django.contrib.auth.models.User):
                The user.

        Returns:
            list of unicode:
            The uniquely identifying information for the user's avatar.
        """
        urls = self._settings_manager_class(user).configuration_for(
            self.avatar_service_id)

        data = [self.avatar_service_id]

        # The sort order here does not matter as long as it is consistent.
        for key in sorted(urls):
            data.append(key)
            data.append(urls[key])

        return data



from django.conf import settings
from django.contrib.sites.models import Site
from django.core.exceptions import ImproperlyConfigured

from djblets.features.checkers import (BaseFeatureChecker,
                                       SettingsFeatureChecker,
                                       SiteConfigFeatureChecker,
                                       get_feature_checker,
                                       set_feature_checker)
from djblets.features.level import FeatureLevel
from djblets.siteconfig.models import SiteConfiguration
from djblets.testing.testcases import TestCase


class BadFeatureChecker(BaseFeatureChecker):
    def __init__(self):
        raise Exception('oh no what even happened')


class FeatureCheckerTests(TestCase):
    """Unit tests for djblets.features.checkers."""

    def tearDown(self):
        super(FeatureCheckerTests, self).tearDown()

        set_feature_checker(None)

        delattr(settings, 'FEATURE_CHECKER')

    def test_get_feature_checker(self):
        """Testing get_feature_checker"""
        set_feature_checker(None)

        settings.FEATURE_CHECKER = \
            'djblets.features.checkers.SiteConfigFeatureChecker'

        self.assertEqual(get_feature_checker().__class__,
                         SiteConfigFeatureChecker)

    def test_get_feature_checker_with_invalid_path(self):
        """Testing get_feature_checker with invalid feature checker path"""
        set_feature_checker(None)

        settings.FEATURE_CHECKER = \
            'djblets.features.checkers.SomethingNotHere'

        with self.assertRaises(ImproperlyConfigured):
            get_feature_checker()

    def test_get_feature_checker_with_init_error(self):
        """Testing get_feature_checker with initialization error"""
        set_feature_checker(None)

        settings.FEATURE_CHECKER = \
            'djblets.features.tests.test_checkers.BadFeatureChecker'

        with self.assertRaises(ImproperlyConfigured):
            get_feature_checker()


class BaseFeatureCheckerTests(TestCase):
    """Unit tests for djblets.features.checkers.BaseFeatureChecker."""

    def setUp(self):
        super(BaseFeatureCheckerTests, self).setUp()

        self.checker = BaseFeatureChecker()
        self._old_debug = settings.DEBUG

    def tearDown(self):
        super(BaseFeatureCheckerTests, self).tearDown()

        self.checker = BaseFeatureChecker()
        settings.DEBUG = self._old_debug

        try:
            delattr(settings, 'MIN_ENABLED_FEATURE_LEVEL')
        except AttributeError:
            pass

    def test_min_enabled_level_with_debug_true(self):
        """Testing BaseFeatureChecker.min_enabled_level with DEBUG=True"""
        settings.DEBUG = True
        self.assertEqual(self.checker.min_enabled_level, FeatureLevel.BETA)

    def test_min_enabled_level_with_debug_false(self):
        """Testing BaseFeatureChecker.min_enabled_level with DEBUG=False"""
        settings.DEBUG = False
        self.assertEqual(self.checker.min_enabled_level, FeatureLevel.STABLE)

    def test_min_enabled_level_with_setting(self):
        """Testing BaseFeatureChecker.min_enabled_level with
        MIN_ENABLED_FEATURE_LEVEL setting
        """
        settings.MIN_ENABLED_FEATURE_LEVEL = 42
        self.assertEqual(self.checker.min_enabled_level, 42)


class SettingsFeatureCheckerTests(TestCase):
    """Unit tests for djblets.features.checkers.SettingsFeatureChecker."""

    def setUp(self):
        super(SettingsFeatureCheckerTests, self).setUp()

        self.checker = SettingsFeatureChecker()

    def tearDown(self):
        super(SettingsFeatureCheckerTests, self).tearDown()

        try:
            delattr(settings, 'ENABLED_FEATURES')
        except AttributeError:
            pass

    def test_is_feature_enabled_with_enabled(self):
        """Testing SettingsFeatureChecker.is_feature_enabled with feature
        enabled
        """
        settings.ENABLED_FEATURES = {
            'my-feature': True,
        }

        self.assertTrue(self.checker.is_feature_enabled('my-feature'))

    def test_is_feature_enabled_with_disabled(self):
        """Testing SettingsFeatureChecker.is_feature_enabled with feature
        disabled
        """
        settings.ENABLED_FEATURES = {
            'my-feature': False,
        }

        self.assertFalse(self.checker.is_feature_enabled('my-feature'))

    def test_is_feature_enabled_with_unset(self):
        """Testing SettingsFeatureChecker.is_feature_enabled with feature state
        not set
        """
        self.assertFalse(self.checker.is_feature_enabled('my-feature'))


class SiteConfigFeatureCheckerTests(TestCase):
    """Unit tests for djblets.features.checkers.SiteConfigFeatureChecker."""

    def setUp(self):
        super(SiteConfigFeatureCheckerTests, self).setUp()

        self.checker = SiteConfigFeatureChecker()
        self.siteconfig = SiteConfiguration.objects.create(
            site=Site.objects.get_current())

    def tearDown(self):
        super(SiteConfigFeatureCheckerTests, self).tearDown()

        try:
            delattr(settings, 'ENABLED_FEATURES')
        except AttributeError:
            pass

    def test_is_feature_enabled_with_siteconfig_enabled(self):
        """Testing SiteConfigFeatureChecker.is_feature_enabled with feature
        enabled in SiteConfiguration
        """
        self.siteconfig.set('enabled_features', {
            'my-feature': True,
        })
        self.siteconfig.save()

        settings.ENABLED_FEATURES = {
            'my-feature': False,
        }

        self.assertTrue(self.checker.is_feature_enabled('my-feature'))

    def test_is_feature_enabled_with_siteconfig_disabled(self):
        """Testing SiteConfigFeatureChecker.is_feature_enabled with feature
        disabled in SiteConfiguration
        """
        self.siteconfig.set('enabled_features', {
            'my-feature': False,
        })
        self.siteconfig.save()

        settings.ENABLED_FEATURES = {
            'my-feature': True,
        }

        self.assertFalse(self.checker.is_feature_enabled('my-feature'))

    def test_is_feature_enabled_with_settings_enabled(self):
        """Testing SiteConfigFeatureChecker.is_feature_enabled with feature
        enabled in Settings
        """
        settings.ENABLED_FEATURES = {
            'my-feature': True,
        }

        self.assertTrue(self.checker.is_feature_enabled('my-feature'))

    def test_is_feature_enabled_with_settings_disabled(self):
        """Testing SiteConfigFeatureChecker.is_feature_enabled with feature
        disabled in Settings
        """
        settings.ENABLED_FEATURES = {
            'my-feature': False,
        }

        self.assertFalse(self.checker.is_feature_enabled('my-feature'))

    def test_is_feature_enabled_with_unset(self):
        """Testing SiteConfigFeatureChecker.is_feature_enabled with feature
        state not set
        """
        self.assertFalse(self.checker.is_feature_enabled('my-feature'))



"""Representations of field types in the API."""

import json
from datetime import datetime
from importlib import import_module

import dateutil.parser
from django.core.exceptions import ValidationError
from django.utils import timezone
from django.utils.translation import gettext_lazy as _, gettext
from pytz.exceptions import AmbiguousTimeError


class BaseAPIFieldType(object):
    """Base class for a field type for an API.

    This is responsible for defining the requirements of a field and to
    validate and normalize a value provided by a client of the API, for storage
    or processing.
    """

    #: The localized name of the field type.
    #:
    #: This should be in sentence casing.
    name = None

    def __init__(self, field_info):
        """Initialize the field type.

        Subclasses should override this if they need to look up any values
        from the field information dictionary.

        Args:
            field_info (dict):
                Information defined for a field.
        """
        self.field_info = field_info

    def get_value_from_data(self, name, fields_data, files_data):
        """Return a value from the data from a request.

        Args:
            name (unicode):
                The name of the entry in the API request to retrieve data for.

            fields_data (dict):
                A dictionary of data provided in the request for standard
                fields. This is usually ``request.POST`` or ``request.GET``.

            files_data (dict):
                A dictionary of data provided in the request for uploaded
                files. This is usually ``request.FILES``.

        Returns:
            object:
            The value from one of the dictionaries, or ``None`` if not found.
        """
        return fields_data.get(name)

    def get_field_info_key(self, key):
        """Return the value for a key in the field information dictionary.

        This will return a consistent error if the key is not found.

        Args:
            key (unicode):
                The name of the key.

        Returns:
            object:
            The value for the key.

        Raises:
            KeyError:
                The key was not found in the field information dictionary.
        """
        try:
            return self.field_info[key]
        except KeyError:
            raise KeyError(
                gettext('Missing "%(key)s" key in %(field_info)r')
                % {
                    'key': key,
                    'field_info': self.field_info,
                })

    def clean_value(self, value):
        """Validate and return a normalized result from the given value.

        By default, this just returns the provided value. Subclasses should
        override this to perform normalization.

        Args:
            value (object):
                The value to validate and normalize.

        Returns:
            object:
            The normalized value.

        Raises:
            django.core.exceptions.ValidationError:
                The value is not valid for this field type.
        """
        return value

    def __str__(self):
        """Return a string representation of the field type.

        This should return a string summary suitable for human consumption.

        Returns:
            unicode:
            A string representation of this field type.
        """
        return str(self.name)


class NonRequestFieldTypeMixin(object):
    """Mixin for field types not intended to handle values from requests."""

    def clean_value(self, value):
        """Validate and return a normalized string result from a value.

        Args:
            value (object):
                The value to normalize.

        Raises:
            NotImplementederror:
                An error describing that this field type cannot be used with
                request data.
        """
        raise NotImplementedError(
            gettext('%s cannot be used for request data')
            % self.__class__.__name__)


class BooleanFieldType(BaseAPIFieldType):
    """A field type for boolean values."""

    name = _('Boolean')

    def clean_value(self, value):
        """Validate and return a boolean result from a value.

        This treats ``1``, ``"1"``, ``True``, or ``"true"`` (case-insensitive)
        as truthy values, and everything else as a falsy value.

        Args:
            value (object):
                The value to normalize.

        Returns:
            unicode:
            The normalized boolean value.
        """
        if isinstance(value, str):
            return value.lower() in ('1', 'true')

        return value in (1, True)


class ChoiceFieldType(BaseAPIFieldType):
    """A field type for a fixed choice of strings.

    This takes a ``choices`` key in the field information dictionary, which
    specifies a list of values accepted during validation.
    """

    name = _('Choice')

    def __init__(self, *args, **kwargs):
        """Initialize the field type.

        Args:
            *args (tuple):
                Positional arguments for the parent constructor.

            **kwargs (dict):
                Keyword arguments for the parent constructor.

        Raises:
            KeyError:
                The "choices" key was not found in the field information.
        """
        super(ChoiceFieldType, self).__init__(*args, **kwargs)

        self.choices = self.get_field_info_key('choices')

    def clean_value(self, value):
        """Validate and return a normalized result from the given value.

        This will return the value if it's a valid choice.

        Args:
            value (object):
                The value to validate and normalize.

        Returns:
            object:
            The value, if it's a valid choice.

        Raises:
            django.core.exceptions.ValidationError:
                The value is not one of the valid choices.
        """
        if value in self.choices:
            return value

        raise ValidationError(
            gettext('"%(value)s" is not a valid value. Valid values '
                    'are: %(choices)s'),
            params={
                'value': value,
                'choices': self._get_choices_str(),
            })

    def __str__(self):
        """Return a string representation of the field type.

        Returns:
            unicode:
            A string representation of this field type.
        """
        return gettext('One of %s') % self._get_choices_str()

    def _get_choices_str(self):
        """Return a string representation of a list of choices.

        This is used for the validation error and string representation.

        Returns:
            unicode:
            A string representing the list of choices.
        """
        return ', '.join(
            '"%s"' % choice
            for choice in self.choices
        )


class DateTimeFieldType(BaseAPIFieldType):
    """A field type for representing date/times in ISO 8601 format."""

    name = _('ISO 8601 Date/Time')

    def clean_value(self, value):
        """Validate and return a datetime from an ISO 8601 value.

        Args:
            value (object):
                The value to validate and normalize. This should be a
                :py:class:`datetime.datetime` or an ISO 8601 date/time string.

        Returns:
            datetime.datetime:
            The resulting date/time value.

        Raises:
            django.core.exceptions.ValidationError:
                The resulting value was not a valid ISO 8601 date/time string
                or the time was ambiguous.
        """
        if not isinstance(value, datetime):
            try:
                value = dateutil.parser.parse(value)
            except ValueError:
                raise ValidationError(
                    gettext('This timestamp is not a valid ISO 8601 '
                            'date/time'))

        if timezone.is_naive(value):
            try:
                value = timezone.make_aware(value,
                                            timezone.get_current_timezone())
            except AmbiguousTimeError:
                raise ValidationError(
                    gettext('This timestamp needs a UTC offset to avoid '
                            'being ambiguous due to daylight savings time '
                            'changes'))

        return value


class DictFieldType(BaseAPIFieldType):
    """A field type for dictionary-based values."""

    name = _('Dictionary')

    def clean_value(self, value):
        """Validate and return a dictionary from a dictionary or JSON value.

        Args:
            value (object):
                The value to validate and normalize. This should be a
                :py:class:`dict` or JSON string representing a dictionary.

        Returns:
            dict:
            The resulting dictionary value.

        Raises:
            django.core.exceptions.ValidationError:
                The resulting value was not a dictionary. Either the value
                provided was of an invalid type or the parsed JSON data did
                not result in a dictionary.
        """
        if isinstance(value, dict):
            return value
        elif isinstance(value, str):
            try:
                result = json.loads(value)
            except ValueError:
                raise ValidationError(
                    gettext('This value is not a valid JSON document'))

            if isinstance(result, dict):
                return result

        raise ValidationError(
            gettext('This value is not a valid dictionary value'))


class FileFieldType(BaseAPIFieldType):
    """A field type for uploaded files."""

    name = _('Uploaded file')

    def get_value_from_data(self, name, fields_data, files_data):
        """Return a value from the uploaded files from a request.

        Args:
            name (unicode):
                The name of the entry in the API request to retrieve data for.

            fields_data (dict, unused):
                A dictionary of data provided in the request for standard
                fields. This is usually ``request.POST`` or ``request.GET``.

            files_data (dict):
                A dictionary of data provided in the request for uploaded
                files. This is usually ``request.FILES``.

        Returns:
            django.core.files.uploadedfile.UploadedFile:
            The value from ``files_data``, or ``None`` if not found.
        """
        return files_data.get(name)


class IntFieldType(BaseAPIFieldType):
    """A field type for integer values."""

    name = _('Integer')

    def clean_value(self, value):
        """Validate and return an integer for a given value.

        Args:
            value (object):
                The value to validate and normalize.

        Returns:
            int:
            The resulting integer value.

        Raises:
            django.core.exceptions.ValidationError:
                The value is not a valid integer.
        """
        try:
            return int(value)
        except ValueError:
            raise ValidationError('"%s" is not an integer' % value)


class ListFieldType(BaseAPIFieldType):
    """A field type for list-based values.

    This takes an optional ``items`` key in the field information dictionary,
    which is itself a field information dictionary for the type of item in the
    list. If provided, all values in a JSON list in the request payload will be
    cleaned using that item type. If not provided, the list will be allowed
    as-is.
    """

    name = _('List')

    def __init__(self, *args, **kwargs):
        """Initialize the field type.

        Args:
            *args (tuple):
                Positional arguments for the parent constructor.

            **kwargs (dict):
                Keyword arguments for the parent constructor.
        """
        super(ListFieldType, self).__init__(*args, **kwargs)

        self.item_info = self.field_info.get('items')

    def clean_value(self, value):
        """Validate and return a list from a list or JSON value.

        Args:
            value (object):
                The value to validate and normalize. This should be a
                :py:class:`list` or JSON string representing a list.

        Returns:
            list:
            The resulting list of optionally-cleaned items.

        Raises:
            django.core.exceptions.ValidationError:
                This will be raised if the resulting value was not a
                dictionary. Either the value provided was of an invalid type or
                the parsed JSON data did not result in a dictionary.

                It may also be raised through the list item type's validation.
        """
        if isinstance(value, str):
            try:
                value = json.loads(value)
            except ValueError:
                raise ValidationError(
                    gettext('This value is not a valid JSON document'))

        if not isinstance(value, list):
            raise ValidationError(gettext('This value is not a valid list'))

        if self.item_info:
            item_type = self.item_info['type'](self.item_info)

            value = [
                item_type.clean_value(item_value)
                for item_value in value
            ]

        return value

    def __str__(self):
        """Return a string representation of the field type.

        Returns:
            unicode:
            A string representation of this field type.
        """
        if self.item_info:
            item_type = self.item_info['type'](self.item_info)

            return gettext('List of %s') % item_type
        else:
            return gettext('List')


class ResourceFieldType(NonRequestFieldTypeMixin, BaseAPIFieldType):
    """A field type for referencing a resource.

    This is intended purely for the :py:class:`Resource.fields
    <djblets.webapi.resources.base.WebAPIResource.fields>` list, for the
    purpose of documentation. It's generally not considered safe to let a
    client specify information like resources in a request. If needed,
    subclasses can override :py:meth:`clean_value` to safely handle client
    requests.

    This expects a "resource" key in the field information.  It can point to
    a resource class or instance, or a string path pointing to one.
    """

    name = _('Resource')

    def __init__(self, *args, **kwargs):
        """Initialize the field type.

        Args:
            *args (tuple):
                Positional arguments for the parent constructor.

            **kwargs (dict):
                Keyword arguments for the parent constructor.

        Raises:
            ImportError:
                The resource class could not be imported.

            KeyError:
                The "resource" key was not found in the field information.
        """
        super(ResourceFieldType, self).__init__(*args, **kwargs)

        resource = self.get_field_info_key('resource')

        if isinstance(resource, str):
            module, attr = resource.rsplit('.', 1)

            try:
                resource = getattr(import_module(module), attr)
            except (AttributeError, ImportError) as e:
                raise ImportError('Unable to load resource "%s": %s'
                                  % (resource, e))

        self.resource = resource

    def __str__(self):
        """Return a string representation of the field type.

        Returns:
            unicode:
            A string representation of this field type.
        """
        return self.resource.__name__


class ResourceListFieldType(ResourceFieldType):
    """A field type for referencing a list of a type of resource.

    This is intended purely for the :py:class:`Resource.fields
    <djblets.webapi.resources.base.WebAPIResource.fields>` list, for the
    purpose of documentation. It's generally not considered safe to let a
    client specify information like resources in a request. If needed,
    subclasses can override :py:meth:`clean_value` to safely handle client
    requests.

    This expects a ``"resource"`` key in the field information.
    """

    name = _('Resource List')

    def __str__(self):
        """Return a string representation of the field type.

        Returns:
            unicode:
            A string representation of this field type.
        """
        return gettext('List of %s') % self.resource.__name__


class StringFieldType(BaseAPIFieldType):
    """A field type for string values."""

    name = _('String')

    def clean_value(self, value):
        """Validate and return a normalized string result from a value.

        Args:
            value (object):
                The value to normalize.

        Returns:
            unicode:
            A string version of the value.
        """
        if isinstance(value, bytes):
            return value.decode('utf-8')
        else:
            return str(value)



"""Internal support for handling deprecations in Djblets.

The version-specific objects in this module are not considered stable between
releases, and may be removed at any point. The base objects are considered
stable.
"""

import warnings

from django.utils.functional import SimpleLazyObject


class BaseRemovedInDjbletsVersionWarning(DeprecationWarning):
    """Base class for a Djblets deprecation warning.

    All version-specific deprecation warnings inherit from this, allowing
    callers to check for Djblets deprecations without being tied to a specific
    version.
    """

    @classmethod
    def warn(cls, message, stacklevel=2):
        """Emit the deprecation warning.

        This is a convenience function that emits a deprecation warning using
        this class, with a suitable default stack level. Callers can provide
        a useful message and a custom stack level.

        Version Added:
            3.0

        Args:
            message (str):
                The message to show in the deprecation warning.

            stacklevel (int, optional):
                The stack level for the warning.
        """
        warnings.warn(message, cls, stacklevel=stacklevel + 1)


class RemovedInDjblets40Warning(BaseRemovedInDjbletsVersionWarning):
    """Deprecations for features scheduled for removal in Djblets 4.0.

    Note that this class will itself be removed in Djblets 4.0. If you need to
    check against Djblets deprecation warnings, please see
    :py:class:`BaseRemovedInDjbletsVersionWarning`. Alternatively, you can use
    the alias for this class, :py:data:`RemovedInNextDjbletsVersionWarning`.
    """


class RemovedInDjblets50Warning(BaseRemovedInDjbletsVersionWarning):
    """Deprecations for features scheduled for removal in Djblets 5.0.

    Note that this class will itself be removed in Djblets 5.0. If you need to
    check against Djblets deprecation warnings, please see
    :py:class:`BaseRemovedInDjbletsVersionWarning`. Alternatively, you can use
    the alias for this class, :py:data:`RemovedInNextDjbletsVersionWarning`.
    """


#: An alias for the next release of Djblets where features would be removed.
RemovedInNextDjbletsVersionWarning = RemovedInDjblets40Warning


def deprecated_arg_value(owner_name, value, old_arg_name, new_arg_name=None,
                         warning_cls=DeprecationWarning):
    """Wrap a value in a lazy object to warn when used.

    Args:
        owner_name (unicode):
            The name of the owner of this argument.

        value (object):
            The argument value.

        old_arg_name (unicode):
            The name of the argument that was deprecated.

        new_arg_name (unicode, optional):
            The optional name of the argument to use in the deprecated
            argument's place, if one is available.

        warning_cls (type, optional):
            The class to use for the warning. This should be
            :py:exc:`DeprecationWarning`, :py:exc:`PendingDeprecationWarning`,
            or a subclass of one.

    Returns:
        django.utils.functional.SimpleLazyObject:
        The value wrapped in a lazy object. The first time it is accessed,
        a warning will be emitted.
    """
    def _warn_on_use():
        if new_arg_name:
            warnings.warn('The "%s" argument for "%s" has been deprecated '
                          'and will be removed in a future version. Use "%s" '
                          'instead.'
                          % (old_arg_name, owner_name, new_arg_name),
                          warning_cls)
        else:
            warnings.warn('The "%s" argument for "%s" has been deprecated '
                          'and will be removed in a future version.'
                          % (old_arg_name, owner_name),
                          warning_cls)

        return value

    return SimpleLazyObject(_warn_on_use)



"""Extension models."""

from django.db import models

from djblets.db.fields import JSONField
from djblets.extensions.errors import InvalidExtensionError


class RegisteredExtension(models.Model):
    """Extension registration info.

    An extension that was both installed and enabled at least once. This
    may contain settings for the extension.

    This does not contain full information for the extension, such as the
    author or description. That is provided by the Extension object itself.
    """
    class_name = models.CharField(max_length=128, unique=True)
    name = models.CharField(max_length=32)
    enabled = models.BooleanField(default=False)
    installed = models.BooleanField(default=False)
    settings = JSONField()

    def __str__(self):
        return self.name

    def get_extension_class(self):
        """Retrieves the python object for the extensions class."""
        if not hasattr(self, '_extension_class'):
            cls = None

            try:
                # Import the function here to avoid a mutual
                # dependency.
                from djblets.extensions.manager import get_extension_managers

                for manager in get_extension_managers():
                    try:
                        cls = manager.get_installed_extension(self.class_name)
                        break
                    except InvalidExtensionError:
                        continue
            except Exception:
                return None

            self._extension_class = cls

        return self._extension_class

    extension_class = property(get_extension_class)

    class Meta:
        # Djblets 0.9+ sets an app label of "djblets_extensions" on
        # Django 1.7+, which would affect the table name. We need to retain
        # the old name for backwards-compatibility.
        db_table = 'extensions_registeredextension'



"""Extension-related signals."""

from django.dispatch import Signal


#: A signal fired when an extension is disabled.
#:
#: Args:
#:     extension (djblets.extensions.extension.Extension):
#:         The extension that was disabled.
extension_disabled = Signal()


#: A signal fired when an extension is enabled.
#:
#: Args:
#:     extension (djblets.extensions.extension.Extension):
#:         The extension that was enabled.
extension_enabled = Signal()


#: A signal fired when an extension is initialized.
#:
#: Args:
#:     extension (type):
#:         The extension class that was initialized.
#:
#:         This will be a subclass of
#:         :py:class:`djblets.extensions.extension.Extension`.
extension_initialized = Signal()


#: A signal fired when an extension is uninitialized.
#:
#: Args:
#:     extension (type):
#:         The extension class that was uninitialized.
#:
#:         This will be a subclass of
#:         :py:class:`djblets.extensions.extension.Extension`.
extension_uninitialized = Signal()


#: A signal fired when an extension's settings are saved.
settings_saved = Signal()



"""Template context processors for SiteConfiguration information."""

import logging

from django.conf import settings

from djblets.siteconfig.models import SiteConfiguration


logger = logging.getLogger(__name__)


def siteconfig(request):
    """Return context variables for accessing site configuration data.

    This will provide templates with a ``siteconfig`` variable, representing
    the current :py:class:`~djblets.siteconfig.models.SiteConfiguration` for
    the installation, and ``siteconfig_settings``, representing all settings on
    the current configuration.

    ``siteconfig_settings`` is preferred over accessing
    :py:attr:`siteconfig.settings
    <djblets.siteconfig.models.SiteConfiguration.settings>`, as it will
    properly handle returning default values.

    Args:
        request (django.http.HttpRequest):
            The current HTTP request.

    Returns:
        dict:
        A dictionary with ``siteconfig`` and ``siteconfig_settings`` keys.
    """
    try:
        siteconfig = SiteConfiguration.objects.get_current()
        siteconfig_settings = siteconfig.settings_wrapper
    except Exception as e:
        logger.exception('Unable to load SiteConfiguration: %s', e)

        siteconfig = None
        siteconfig_settings = None

    return {
        'siteconfig': siteconfig,
        'siteconfig_settings': siteconfig_settings,
    }


def settings_vars(request):
    """Return context variables for accessing Django settings.

    This will provide templates with a ``settings`` variable, which represents
    the Django settings object.

    Args:
        request (django.http.HttpRequest):
            The current HTTP request.

    Returns:
        dict:
        A dictionary with a ``settings`` key.
    """
    return {'settings': settings}



"""Unit tests for WebAPI decorators."""

from django.contrib.auth.models import AnonymousUser, User
from django.core.files.uploadedfile import SimpleUploadedFile
from django.test.client import RequestFactory

from djblets.testing.testcases import TestCase
from djblets.util.decorators import augment_method_from
from djblets.webapi.decorators import (copy_webapi_decorator_data,
                                       webapi_login_required,
                                       webapi_permission_required,
                                       webapi_request_fields,
                                       webapi_response_errors)
from djblets.webapi.errors import (DOES_NOT_EXIST, INVALID_FORM_DATA,
                                   NOT_LOGGED_IN, PERMISSION_DENIED)
from djblets.webapi.fields import (BooleanFieldType,
                                   ChoiceFieldType,
                                   IntFieldType,
                                   FileFieldType,
                                   StringFieldType)


class WebAPIDecoratorTests(TestCase):
    """Unit tests for djblets.webapi.decorators."""

    def test_copy_webapi_decorator_data(self):
        """Testing copy_webapi_decorator_data"""
        def func1():
            """Function 1"""

        def func2():
            """Function 2"""

        func1.test1 = True
        func1.response_errors = set(['a', 'b'])
        func2.test2 = True
        func2.response_errors = set(['c', 'd'])

        result = copy_webapi_decorator_data(func1, func2)
        self.assertEqual(result, func2)

        self.assertTrue(hasattr(func2, 'test1'))
        self.assertTrue(hasattr(func2, 'test2'))
        self.assertTrue(hasattr(func2, 'response_errors'))
        self.assertTrue(func2.test1)
        self.assertTrue(func2.test2)
        self.assertEqual(func2.response_errors, set(['a', 'b', 'c', 'd']))
        self.assertEqual(func2.__doc__, 'Function 1')
        self.assertEqual(func2.__name__, 'func1')

        self.assertFalse(hasattr(func1, 'test2'))
        self.assertEqual(func1.response_errors, set(['a', 'b']))

    def test_webapi_response_errors_state(self):
        """Testing @webapi_response_errors state"""
        def orig_func():
            """Function 1"""

        func = webapi_response_errors(DOES_NOT_EXIST, NOT_LOGGED_IN)(orig_func)

        self.assertFalse(hasattr(orig_func, 'response_errors'))

        self.assertEqual(func.__name__, 'orig_func')
        self.assertEqual(func.__doc__, 'Function 1')
        self.assertTrue(hasattr(func, 'response_errors'))
        self.assertEqual(func.response_errors,
                         set([DOES_NOT_EXIST, NOT_LOGGED_IN]))

    def test_webapi_response_errors_preserves_state(self):
        """Testing @webapi_response_errors preserves decorator state"""
        @webapi_response_errors(DOES_NOT_EXIST)
        @webapi_response_errors(NOT_LOGGED_IN)
        def func():
            """Function 1"""

        self.assertEqual(func.__name__, 'func')
        self.assertEqual(func.__doc__, 'Function 1')
        self.assertTrue(hasattr(func, 'response_errors'))
        self.assertEqual(func.response_errors,
                         set([DOES_NOT_EXIST, NOT_LOGGED_IN]))

    def test_webapi_response_errors_call(self):
        """Testing @webapi_response_errors calls original function"""
        @webapi_response_errors(DOES_NOT_EXIST, NOT_LOGGED_IN)
        def func():
            func.seen = True

        func()

        self.assertTrue(hasattr(func, 'seen'))

    def test_webapi_login_required_state(self):
        """Testing @webapi_login_required state"""
        def orig_func():
            """Function 1"""

        func = webapi_login_required(orig_func)

        self.assertFalse(hasattr(orig_func, 'login_required'))
        self.assertFalse(hasattr(orig_func, 'response_errors'))

        self.assertEqual(func.__name__, 'orig_func')
        self.assertEqual(func.__doc__, 'Function 1')
        self.assertTrue(hasattr(func, 'response_errors'))
        self.assertTrue(hasattr(func, 'login_required'))
        self.assertTrue(func.login_required)
        self.assertEqual(func.response_errors, set([NOT_LOGGED_IN]))

    def test_webapi_login_required_preserves_state(self):
        """Testing @webapi_login_required preserves decorator state"""
        @webapi_response_errors(DOES_NOT_EXIST)
        def orig_func():
            """Function 1"""

        func = webapi_login_required(orig_func)

        self.assertFalse(hasattr(orig_func, 'login_required'))

        self.assertEqual(func.__name__, 'orig_func')
        self.assertEqual(func.__doc__, 'Function 1')
        self.assertTrue(hasattr(func, 'response_errors'))
        self.assertTrue(hasattr(func, 'login_required'))
        self.assertTrue(func.login_required)
        self.assertEqual(func.response_errors,
                         set([DOES_NOT_EXIST, NOT_LOGGED_IN]))

    def test_webapi_login_required_call_when_authenticated(self):
        """Testing @webapi_login_required calls when authenticated"""
        @webapi_login_required
        def func(request):
            func.seen = True

        request = RequestFactory().request()
        request.user = User()
        result = func(request)

        self.assertTrue(hasattr(func, 'seen'))
        self.assertEqual(result, None)

    def test_webapi_login_required_call_when_anonymous(self):
        """Testing @webapi_login_required calls when anonymous"""
        @webapi_login_required
        def func(request):
            func.seen = True

        request = RequestFactory().request()
        request.user = AnonymousUser()
        result = func(request)

        self.assertFalse(hasattr(func, 'seen'))
        self.assertEqual(result, NOT_LOGGED_IN)

    def test_webapi_permission_required_state(self):
        """Testing @webapi_permission_required state"""
        def orig_func():
            """Function 1"""

        func = webapi_permission_required('myperm')(orig_func)

        self.assertFalse(hasattr(orig_func, 'response_errors'))

        self.assertEqual(func.__name__, 'orig_func')
        self.assertEqual(func.__doc__, 'Function 1')
        self.assertTrue(hasattr(func, 'response_errors'))
        self.assertEqual(func.response_errors,
                         set([NOT_LOGGED_IN, PERMISSION_DENIED]))

    def test_webapi_permission_required_preserves_state(self):
        """Testing @webapi_permission_required preserves decorator state"""
        @webapi_response_errors(DOES_NOT_EXIST)
        def orig_func():
            """Function 1"""

        func = webapi_permission_required('myperm')(orig_func)

        self.assertEqual(func.__name__, 'orig_func')
        self.assertEqual(func.__doc__, 'Function 1')
        self.assertTrue(hasattr(func, 'response_errors'))
        self.assertEqual(func.response_errors,
                         set([DOES_NOT_EXIST, NOT_LOGGED_IN,
                              PERMISSION_DENIED]))

    def test_webapi_permission_required_call_when_anonymous(self):
        """Testing @webapi_permission_required calls when anonymous"""
        @webapi_permission_required('foo')
        def func(request):
            func.seen = True

        request = RequestFactory().request()
        request.user = AnonymousUser()
        result = func(request)

        self.assertFalse(hasattr(func, 'seen'))
        self.assertEqual(result, NOT_LOGGED_IN)

    def test_webapi_permission_required_call_when_has_permission(self):
        """Testing @webapi_permission_required calls when has permission"""
        @webapi_permission_required('foo')
        def func(request):
            func.seen = True

        request = RequestFactory().request()
        request.user = User()
        request.user.has_perm = lambda perm: True
        result = func(request)

        self.assertTrue(hasattr(func, 'seen'))
        self.assertEqual(result, None)

    def test_webapi_permission_required_call_when_no_permission(self):
        """Testing @webapi_permission_required calls when no permission"""
        @webapi_permission_required('foo')
        def func(request):
            func.seen = True

        request = RequestFactory().request()
        request.user = User()
        request.user.has_perm = lambda perm: False
        result = func(request)

        self.assertFalse(hasattr(func, 'seen'))
        self.assertEqual(result, PERMISSION_DENIED)

    def test_webapi_request_fields_state(self):
        """Testing @webapi_request_fields state"""
        def orig_func():
            """Function 1"""

        required = {
            'required_param': {
                'type': BooleanFieldType,
                'description': 'Required param'
            },
        }

        optional = {
            'optional_param': {
                'type': BooleanFieldType,
                'description': 'Optional param'
            },
        }

        func = webapi_request_fields(required, optional)(orig_func)

        self.assertFalse(hasattr(orig_func, 'required_fields'))
        self.assertFalse(hasattr(orig_func, 'optional_fields'))
        self.assertFalse(hasattr(orig_func, 'response_errors'))

        self.assertEqual(func.__name__, 'orig_func')
        self.assertEqual(func.__doc__, 'Function 1')
        self.assertTrue(hasattr(func, 'response_errors'))
        self.assertTrue(hasattr(func, 'required_fields'))
        self.assertTrue(hasattr(func, 'optional_fields'))
        self.assertEqual(func.required_fields, required)
        self.assertEqual(func.optional_fields, optional)
        self.assertEqual(func.response_errors, set([INVALID_FORM_DATA]))

    def test_webapi_request_fields_preserves_state(self):
        """Testing @webapi_request_fields preserves decorator state"""
        required1 = {
            'required1': {
                'type': BooleanFieldType,
                'description': 'Required param'
            },
        }

        optional1 = {
            'optional1': {
                'type': BooleanFieldType,
                'description': 'Optional param'
            },
        }

        @webapi_request_fields(required1, optional1)
        @webapi_response_errors(DOES_NOT_EXIST)
        def orig_func():
            """Function 1"""

        required2 = {
            'required2': {
                'type': BooleanFieldType,
                'description': 'Required param'
            },
        }

        optional2 = {
            'optional2': {
                'type': BooleanFieldType,
                'description': 'Optional param'
            },
        }

        func = webapi_request_fields(required2, optional2)(orig_func)

        expected_required = required1.copy()
        expected_required.update(required2)
        expected_optional = optional1.copy()
        expected_optional.update(optional2)

        self.assertTrue(hasattr(orig_func, 'required_fields'))
        self.assertTrue(hasattr(orig_func, 'optional_fields'))
        self.assertTrue(hasattr(orig_func, 'response_errors'))

        self.assertEqual(func.__name__, 'orig_func')
        self.assertEqual(func.__doc__, 'Function 1')
        self.assertTrue(hasattr(func, 'response_errors'))
        self.assertTrue(hasattr(func, 'required_fields'))
        self.assertTrue(hasattr(func, 'optional_fields'))
        self.assertEqual(func.required_fields, expected_required)
        self.assertEqual(func.optional_fields, expected_optional)
        self.assertEqual(func.response_errors,
                         set([DOES_NOT_EXIST, INVALID_FORM_DATA]))

    def test_webapi_request_fields_checks_inheritance(self):
        """Testing @webapi_request_fields properly checks inherited fields"""
        class A(object):
            @webapi_request_fields(required={
                'required1': {
                    'type': BooleanFieldType,
                    'description': 'Required param',
                },
            })
            def test(*args, **kwargs):
                pass

        class B(A):
            @webapi_request_fields(required={
                'required2': {
                    'type': BooleanFieldType,
                    'description': 'Required param',
                },
            })
            @augment_method_from(A)
            def test(*args, **kwargs):
                pass

        request = RequestFactory().post('/', data={'required2': True})

        b = B()
        error, response = b.test(request)

        self.assertTrue('fields' in response)
        self.assertTrue('required1' in response['fields'])
        self.assertEqual(response['fields']['required1'],
                         ['This field is required'])
        self.assertFalse('required2' in response['fields'])

    def test_webapi_request_fields_call_normalizes_params(self):
        """Testing @webapi_request_fields normalizes params to function"""
        if not hasattr(__builtins__, 'file'):
            # Python 3.x doesn't have `file`, so define our own for the test.
            class file(object):
                pass

        @webapi_request_fields(
            required={
                'required_int': {
                    'type': IntFieldType,
                },
                'required_string': {
                    'type': StringFieldType,
                },
                'required_choice': {
                    'type': ChoiceFieldType,
                    'choices': ['a', 'b', 'c'],
                },
                'required_bool': {
                    'type': BooleanFieldType,
                },
                'required_file': {
                    'type': FileFieldType,
                },
            },
            optional={
                'optional_param': {
                    'type': BooleanFieldType,
                }
            },
        )
        def func(request, required_int=None, required_string=None,
                 required_choice=None, required_file=None, required_bool=None,
                 optional_param=None, parsed_request_fields=None,
                 extra_fields={}):
            func.seen = True
            self.assertEqual(required_int, 42)
            self.assertEqual(required_string, 'test')
            self.assertEqual(required_choice, 'a')
            self.assertEqual(required_file.read(), b'content')
            self.assertTrue(required_bool)
            self.assertTrue(optional_param)
            self.assertFalse(extra_fields)
            self.assertEqual(
                parsed_request_fields,
                {
                    'required_int': required_int,
                    'required_string': required_string,
                    'required_choice': required_choice,
                    'required_bool': required_bool,
                    'required_file': required_file,
                    'optional_param': optional_param,
                })

        result = func(RequestFactory().post(
            path='/',
            data={
                'required_int': '42',
                'required_string': 'test',
                'required_choice': 'a',
                'required_bool': 'true',
                'required_file': SimpleUploadedFile(name='test',
                                                    content=b'content'),
                'optional_param': '1',
            },
        ))

        self.assertTrue(hasattr(func, 'seen'))
        self.assertIsNone(result)

    def test_webapi_request_fields_call_normalizes_params_legacy_types(self):
        """Testing @webapi_request_fields normalizes params to function when
        using legacy types
        """
        if not hasattr(__builtins__, 'file'):
            # Python 3.x doesn't have `file`, so define our own for the test.
            class file(object):
                pass

        @webapi_request_fields(
            required={
                'required_int': {
                    'type': int,
                },
                'required_unicode_str': {
                    'type': str,
                },
                'required_native_str': {
                    'type': str,
                },
                'required_choice_list': {
                    'type': ['a', 'b', 'c'],
                },
                'required_choice_tuple': {
                    'type': ('a', 'b', 'c'),
                },
                'required_bool': {
                    'type': bool,
                },
                'required_file': {
                    'type': file,
                },
            },
            optional={
                'optional_param': {
                    'type': bool,
                }
            },
        )
        def func(request, required_int=None, required_unicode_str=None,
                 required_native_str=None, required_choice_list=None,
                 required_choice_tuple=None, required_file=None,
                 required_bool=None, optional_param=None,
                 parsed_request_fields=None, extra_fields={}):
            func.seen = True
            self.assertEqual(required_int, 42)
            self.assertEqual(required_unicode_str, 'test')
            self.assertEqual(required_native_str, str('test'))
            self.assertEqual(required_choice_list, 'a')
            self.assertEqual(required_choice_tuple, 'c')
            self.assertEqual(required_file.read(), b'content')
            self.assertTrue(required_bool)
            self.assertTrue(optional_param)
            self.assertFalse(extra_fields)
            self.assertEqual(
                parsed_request_fields,
                {
                    'required_int': required_int,
                    'required_unicode_str': required_unicode_str,
                    'required_native_str': required_native_str,
                    'required_choice_list': required_choice_list,
                    'required_choice_tuple': required_choice_tuple,
                    'required_bool': required_bool,
                    'required_file': required_file,
                    'optional_param': optional_param,
                })

        result = func(RequestFactory().post(
            path='/',
            data={
                'required_int': '42',
                'required_unicode_str': 'test',
                'required_native_str': str('test'),
                'required_choice_list': 'a',
                'required_choice_tuple': 'c',
                'required_bool': 'true',
                'required_file': SimpleUploadedFile(name='test',
                                                    content=b'content'),
                'optional_param': '1',
            },
        ))

        self.assertTrue(hasattr(func, 'seen'))
        self.assertIsNone(result)

        self.assertEqual(
            func.required_fields,
            {
                'required_int': {
                    'type': IntFieldType,
                },
                'required_unicode_str': {
                    'type': StringFieldType,
                },
                'required_native_str': {
                    'type': StringFieldType,
                },
                'required_choice_list': {
                    'type': ChoiceFieldType,
                    'choices': ('a', 'b', 'c'),
                },
                'required_choice_tuple': {
                    'type': ChoiceFieldType,
                    'choices': ('a', 'b', 'c'),
                },
                'required_bool': {
                    'type': BooleanFieldType,
                },
                'required_file': {
                    'type': FileFieldType,
                },
            })

    def test_webapi_request_fields_call_with_unexpected_arg(self):
        """Testing @webapi_request_fields with unexpected argument"""
        @webapi_request_fields(
            required={
                'required_param': {
                    'type': IntFieldType,
                }
            },
        )
        def func(request, required_param=None, extra_fields={}):
            func.seen = True

        result = func(RequestFactory().get(
            path='/',
            data={
                'required_param': '42',
                'optional_param': '1',
            }
        ))

        self.assertFalse(hasattr(func, 'seen'))
        self.assertEqual(result[0], INVALID_FORM_DATA)
        self.assertTrue('fields' in result[1])
        self.assertTrue('optional_param' in result[1]['fields'])

    def test_webapi_request_fields_call_with_allow_unknown(self):
        """Testing @webapi_request_fields with allow_unknown=True"""
        @webapi_request_fields(
            required={
                'required_param': {
                    'type': IntFieldType,
                }
            },
            allow_unknown=True
        )
        def func(request, required_param=None, parsed_request_fields=None,
                 extra_fields={}):
            func.seen = True
            self.assertEqual(required_param, 42)
            self.assertTrue('optional_param' in extra_fields)
            self.assertEqual(extra_fields['optional_param'], '1')
            self.assertEqual(parsed_request_fields,
                             {'required_param': required_param})

        result = func(RequestFactory().get(
            path='/',
            data={
                'required_param': '42',
                'optional_param': '1',
            }
        ))

        self.assertTrue(hasattr(func, 'seen'))
        self.assertEqual(result, None)

    def test_webapi_request_fields_call_filter_special_params(self):
        """Testing @webapi_request_fields filters special params"""
        @webapi_request_fields(
            required={
                'required_param': {
                    'type': IntFieldType,
                }
            },
        )
        def func(request, required_param=None, parsed_request_fields=None,
                 extra_fields={}):
            func.seen = True
            self.assertTrue(isinstance(required_param, int))
            self.assertEqual(required_param, 42)
            self.assertEqual(parsed_request_fields,
                             {'required_param': required_param})
            self.assertFalse(extra_fields)

        result = func(RequestFactory().get(
            path='/',
            data={
                'required_param': '42',
                'api_format': 'json',
            }
        ))

        self.assertTrue(hasattr(func, 'seen'))
        self.assertEqual(result, None)

    def test_webapi_request_fields_call_validation_int(self):
        """Testing @webapi_request_fields with int parameter validation"""
        @webapi_request_fields(
            required={
                'myint': {
                    'type': IntFieldType,
                }
            }
        )
        def func(request, myint=False, parsed_request_fields=None,
                 extra_fields={}):
            func.seen = True

        result = func(RequestFactory().get(
            path='/',
            data={
                'myint': 'abc',
            }
        ))
        self.assertFalse(hasattr(func, 'seen'))
        self.assertEqual(result[0], INVALID_FORM_DATA)
        self.assertTrue('fields' in result[1])
        self.assertTrue('myint' in result[1]['fields'])



"""Registry for token generators.

Version Added:
    3.0
"""

import logging
from typing import Optional, Set

from django.conf import settings
from django.core.exceptions import ImproperlyConfigured

from djblets.registries.registry import Registry
from djblets.registries.mixins import ExceptionFreeGetterMixin
from djblets.secrets.token_generators.base import BaseTokenGenerator
from djblets.secrets.token_generators.legacy_sha1 import \
    LegacySHA1TokenGenerator
from djblets.secrets.token_generators.vendor_checksum import \
    VendorChecksumTokenGenerator


logger = logging.getLogger(__name__)


class TokenGeneratorRegistry(ExceptionFreeGetterMixin[BaseTokenGenerator],
                             Registry[BaseTokenGenerator]):
    """Registry for managing token generators.

    Version Added:
        3.0
    """

    lookup_attrs = ['token_generator_id']

    def get_token_generator(
        self,
        token_generator_id: str,
    ) -> Optional[BaseTokenGenerator]:
        """Return a token generator with the specified ID.

        Args:
            token_generator_id (str):
                The ID of the token generator.

        Returns:
            djblets.secrets.token_generators.BaseTokenGenerator:
            The token generator instance, or ``None`` if not found.
        """
        return self.get_or_none('token_generator_id', token_generator_id)

    def get_default(self) -> BaseTokenGenerator:
        """Return the default token generator.

        The default token generator ID can be set in
        ``settings.DJBLETS_DEFAULT_API_TOKEN_GENERATOR``. If not
        set the default token generator will be :py:class:`~djblets.secrets.
        token_generators.vendor_checksum.VendorChecksumTokenGenerator`.

        Returns:
            djblets.secrets.token_generators.BaseTokenGenerator:
            The default token generator.
        """
        token_generator_id = (
            getattr(settings, 'DJBLETS_DEFAULT_API_TOKEN_GENERATOR', None) or
            VendorChecksumTokenGenerator.token_generator_id
        )

        assert isinstance(token_generator_id, str)

        token_generator = self.get_token_generator(token_generator_id)

        if token_generator is None:
            logger.error('Could not find the %s token generator in the '
                         'registry. Using the vendor_checksum token '
                         'generator instead.', token_generator_id)

            token_generator_id = \
                VendorChecksumTokenGenerator.token_generator_id
            assert token_generator_id

            token_generator = self.get_token_generator(token_generator_id)
            assert token_generator is not None

        return token_generator

    def get_defaults(self):
        """Return the default token generators.

        Returns:
            list of djblets.secrets.token_generators.BaseTokenGenerator:
            The list of default token generators.
        """
        return [
            LegacySHA1TokenGenerator(),
            VendorChecksumTokenGenerator(),
        ]

    def get_deprecated_token_generators(self) -> Set[str]:
        """Return deprecated token generator IDs.

        The set of deprecated token generator IDs can be set in
        ``settings.DJBLETS_DEPRECATED_TOKEN_GENERATORS``. By default
        the set will contain the :py:class:`~djblets.secrets.
        token_generators.legacy_sha1.LegacySHA1TokenGenerator`'s ID.

        Version Added:
            3.1

        Returns:
            set of str:
            A set of deprecated token generator IDs.
        """
        deprecated_token_generators = (
            getattr(settings, 'DJBLETS_DEPRECATED_TOKEN_GENERATORS', None) or
            {LegacySHA1TokenGenerator.token_generator_id}
        )

        if type(deprecated_token_generators) is not set:
            raise ImproperlyConfigured(
                'settings.DJBLETS_DEPRECATED_TOKEN_GENERATORS must be a set'
                'of token generator ID strings.'
            )

        return deprecated_token_generators



"""Managers for Django database models."""

from django.db import models, IntegrityError


class ConcurrencyManager(models.Manager):
    """
    A class designed to work around database concurrency issues.
    """
    def get_or_create(self, **kwargs):
        """
        A wrapper around get_or_create that makes a final attempt to get
        the object if the creation fails.

        This helps with race conditions in the database where, between the
        original get() and the create(), another process created the object,
        causing us to fail. We'll then execute a get().

        This is still prone to race conditions, but they're even more rare.
        A delete() would have to happen before the unexpected create() but
        before the get().
        """
        try:
            return super(ConcurrencyManager, self).get_or_create(**kwargs)
        except IntegrityError:
            kwargs.pop('defaults', None)
            return self.get(**kwargs)



"""Form definitions for extensions."""

from djblets.forms.forms import KeyValueForm


class SettingsForm(KeyValueForm):
    """Settings form for extension configuration.

    A base form for loading/saving settings for an extension. This is meant
    to be overridden by extensions to provide configuration pages. Any fields
    defined by the form will be loaded and saved automatically.
    """

    def __init__(self, extension, *args, **kwargs):
        self.request = kwargs.pop('request', None)
        self.extension = extension
        self.settings = extension.settings

        super(SettingsForm, self).__init__(instance=extension.settings,
                                           *args, **kwargs)

    def set_key_value(self, key, value):
        """Set the value for an extension settings key.

        Args:
            key (unicode):
                The settings key.

            value (object):
                The settings value.
        """
        self.instance.set(key, value)

    def save_instance(self):
        """Save the instance."""
        self.instance.save()



"""Unit tests for djblets.privacy.consent.base.ConsentData."""

from datetime import datetime

from django.utils import timezone

from djblets.privacy.consent import ConsentData
from djblets.privacy.tests.testcases import ConsentTestCase


class ConsentDataTests(ConsentTestCase):
    """Unit tests for djblets.privacy.consent.base.ConsentData."""

    def test_parse_audit_info_with_all_data(self):
        """Testing ConsentData.parse_audit_info with all data"""
        consent_data = ConsentData.parse_audit_info('test-requirement', {
            'granted': True,
            'timestamp': '2018-01-02T13:14:15+00:00',
            'source': 'http://example.com/account/profile/#consent',
            'extra_data': {
                'test': True,
            },
        })

        self.assertEqual(consent_data.requirement_id, 'test-requirement')
        self.assertTrue(consent_data.granted)
        self.assertEqual(consent_data.timestamp,
                         datetime(2018, 1, 2, 13, 14, 15, tzinfo=timezone.utc))
        self.assertEqual(consent_data.source,
                         'http://example.com/account/profile/#consent')
        self.assertEqual(
            consent_data.extra_data,
            {
                'test': True,
            })

    def test_parse_audit_info_with_minimum_data(self):
        """Testing ConsentData.parse_audit_info with minimum required data"""
        consent_data = ConsentData.parse_audit_info('test-requirement', {
            'granted': False,
            'timestamp': '2018-01-02T13:14:15+00:00',
        })

        self.assertEqual(consent_data.requirement_id, 'test-requirement')
        self.assertFalse(consent_data.granted)
        self.assertEqual(consent_data.timestamp,
                         datetime(2018, 1, 2, 13, 14, 15, tzinfo=timezone.utc))
        self.assertIsNone(consent_data.source)
        self.assertIsNone(consent_data.extra_data)

    def test_serialize_audit_info_with_all_data(self):
        """Testing ConsentData.serialize_audit_info with all data"""
        consent_data = ConsentData(
            requirement_id='test-requirement',
            granted=True,
            timestamp=datetime(2018, 1, 2, 13, 14, 15, tzinfo=timezone.utc),
            source='http://example.com/account/profile/#consent',
            extra_data={
                'test': True,
            })

        self.assertEqual(
            consent_data.serialize_audit_info('123:test@example.com'),
            {
                'identifier': '123:test@example.com',
                'granted': True,
                'timestamp': '2018-01-02T13:14:15+00:00',
                'source': 'http://example.com/account/profile/#consent',
                'extra_data': {
                    'test': True,
                },
            })

    def test_serialize_audit_info_with_minimum_data(self):
        """Testing ConsentData.serialize_audit_info with minimum required data
        """
        consent_data = ConsentData(
            requirement_id='test-requirement',
            granted=False,
            timestamp=datetime(2018, 1, 2, 13, 14, 15, tzinfo=timezone.utc))

        self.assertEqual(
            consent_data.serialize_audit_info('123:test@example.com'),
            {
                'identifier': '123:test@example.com',
                'granted': False,
                'timestamp': '2018-01-02T13:14:15+00:00',
            })



"""A base form for working with settings stored on SiteConfigurations."""

from __future__ import annotations

from typing import Optional

from django.http import HttpRequest

from djblets.forms.forms import KeyValueForm
from djblets.siteconfig.models import (SiteConfiguration,
                                       SiteConfigurationSettingsValue)


class SiteSettingsForm(KeyValueForm):
    """A base form for loading/saving settings for a SiteConfiguration.

    This is meant to be subclassed for different settings pages. Any fields
    defined by the form will be loaded/saved automatically.
    """

    #: The HTTP request used for this form.
    #:
    #: Type:
    #:     django.http.HttpRequest
    request: HttpRequest

    #: The site configuration settings that are loaded from and saved to.
    #:
    #: Type:
    #:      djblets.siteconfig.models.SiteConfiguration
    siteconfig: SiteConfiguration

    def __init__(
        self,
        siteconfig: SiteConfiguration,
        *args,
        **kwargs,
    ) -> None:
        """Initialize the form.

        Args:
            siteconfig (djblets.siteconfig.models.SiteConfiguration):
                The site configuration to work with on the form.

            *args (tuple):
                Positional arguments to pass to the parent constructor.

            **kwargs (dict):
                Keyword arguments to pass to the parent constructor.
        """
        self.request = kwargs.pop('request', None)
        self.siteconfig = siteconfig

        super().__init__(instance=siteconfig, *args, **kwargs)

    def get_key_value(
        self,
        key: str,
        default: Optional[SiteConfigurationSettingsValue] = None,
    ) -> Optional[SiteConfigurationSettingsValue]:
        """Return the value for a SiteConfiguration settings key.

        Args:
            key (str):
                The settings key.

            default (object, unused):
                The default value from the form, which will be ignored,
                so that the registered siteconfig defaults will be used.

        Returns:
            object:
            The resulting value from the settings.
        """
        return self.siteconfig.get(key)

    def set_key_value(
        self,
        key: str,
        value: SiteConfigurationSettingsValue,
    ) -> None:
        """Set the value for a SiteConfiguration settings key.

        Args:
            key (str):
                The settings key.

            value (object):
                The settings value.
        """
        self.siteconfig.set(key, value)

    def save_instance(self) -> None:
        """Save the SiteConfiguration instance."""
        self.siteconfig.save()



"""Avatar-specific errors."""

from djblets.registries.errors import ItemLookupError


class AvatarServiceNotFoundError(ItemLookupError):
    """An avatar service specific lookup error."""


class DisabledServiceError(Exception):
    """An exception for using disabled avatar services.

    This will be thrown when a disabled avatar service is set to be the default
    avatar service.
    """



"""Forms, fields, and widgets for gathering and displaying consent."""

from django import forms
from django.contrib import messages
from django.forms.widgets import MultiWidget, Widget
from django.template.loader import render_to_string
from django.utils import timezone
from django.utils.html import format_html_join
from django.utils.translation import gettext_lazy as _

from djblets.privacy.consent import (Consent,
                                     get_consent_requirements_registry,
                                     get_consent_tracker)


class ConsentRequirementWidget(Widget):
    """A widget for displaying and gathering consent information.

    This presents the consent requirement in an easy-to-digest form, and
    makes it easy for users to choose to allow or block the thing requiring
    consent.

    This is meant to be used with :py:class:`ConsentRequirementField`.
    """

    def __init__(self, consent_requirement=None, attrs=None):
        """Initialize the widget.

        Args:
            consent_requirement (djblets.privacy.consent.base.
                                 ConsentRequirement, optional):
                The consent requirement information. If not provided, this
                must be directly set on the instance before rendering the
                widget.

            attrs (dict, optional):
                HTML attributes for the widget. This is used only to set an
                ``id`` attribute for the field.
        """
        super(ConsentRequirementWidget, self).__init__(attrs=attrs)

        self.consent_requirement = consent_requirement

    def render(self, name, value, attrs=None, renderer=None):
        """Render the widget.

        Args:
            name (unicode):
                The base name used for the ``<input>`` elements. Specific
                names will be composed from this.

            value (unicode):
                The current value for the field.

            attrs (dict, optional):
                HTML attributes for the widget. This is used only to set an
                ``id`` attribute for the field.

            renderer (django.forms.renderers.BaseRenderer, optional):
                The form renderer.

        Returns:
            django.utils.safestring.SafeText:
            The rendered HTML for the widget.
        """
        if attrs is None:
            attrs = {}

        return render_to_string(
            'privacy/consent/consent_widget.html',
            context={
                'ALLOW': ConsentRequirementField.ALLOW,
                'BLOCK': ConsentRequirementField.BLOCK,
                'consent_requirement': self.consent_requirement,
                'icons': self.consent_requirement.icons,
                'id': attrs.get('id', ''),
                'name': name,
                'value': value,
            })

    def value_from_datadict(self, data, files, name):
        """Return the field value from the submitted form data.

        Args:
            data (dict):
                The submitted form data.

            files (dict, unused):
                The submitted files data.

            name (unicode):
                The base name for the form fields.

        Returns:
            unicode:
            A value for the fields. This will be one of
            :py:attr:`ConsentRequirementField.ALLOW`,
            :py:attr:`ConsentRequirementField.BLOCK`, or ``None``.
        """
        return data.get('%s_choice' % name)


class MultiConsentRequirementsWidget(MultiWidget):
    """A widget for displaying and gathering multiple consent information.

    This adds a :py:class:`ConsentRequirementWidget` for each consent
    requirement provided to the widget.

    This is meant to be used with :py:class:`MultiConsentRequirementsField`.
    """

    def __init__(self, consent_requirements, attrs=None):
        """Initialize the widget.

        Args:
            consent_requirements (list of djblets.privacy.consent.base.
                                  ConsentRequirement):
                The list of consent requirements.

            attrs (dict, optional):
                HTML attributes for the widget. This is used only to set an
                ``id`` attribute for the field.
        """
        self.consent_requirements = consent_requirements

        super(MultiConsentRequirementsWidget, self).__init__(
            widgets=[
                ConsentRequirementWidget(consent_requirement=requirement)
                for requirement in consent_requirements
            ],
            attrs=attrs)

    def render(self, name, value, attrs=None, renderer=None):
        """Render the widget.

        Args:
            name (unicode):
                The base name used for the ``<input>`` elements. Specific
                names will be composed from this.

            value (list of unicode):
                The current values for the fields.

            attrs (dict, optional):
                HTML attributes for the widget. This is used only to set a
                base ``id`` attribute for the fields.

            renderer (django.forms.renderers.BaseRenderer, optional):
                The form renderer.

        Returns:
            django.utils.safestring.SafeText:
            The rendered HTML for the widget.
        """
        if not isinstance(value, list):
            value = self.decompress(value)

        final_attrs = self.build_attrs(attrs)
        field_id = final_attrs.get('id')
        output = []

        for i, widget in enumerate(self.widgets):
            consent_requirement = widget.consent_requirement
            requirement_id = consent_requirement.requirement_id

            try:
                widget_value = value[i]
            except IndexError:
                widget_value = None

            if field_id:
                final_attrs = dict(final_attrs,
                                   id='%s_%s' % (field_id, requirement_id))

            rendered = widget.render('%s_%s' % (name, requirement_id),
                                     widget_value, final_attrs,
                                     renderer=renderer)
            output.append((rendered,))

        return format_html_join('', '{0}', output)

    def value_from_datadict(self, data, files, name):
        """Return the field values from the submitted form data.

        Args:
            data (dict):
                The submitted form data.

            files (dict, unused):
                The submitted files data.

            name (unicode):
                The base name for the form fields.

        Returns:
            list of unicode:
            A list of values for all the fields, in the order of the list of
            consent requirements provided to the widget. Each item will be
            one of :py:attr:`ConsentRequirementField.ALLOW`,
            :py:attr:`ConsentRequirementField.BLOCK`, or ``None``.
        """
        return [
            widget.value_from_datadict(
                data,
                files,
                '%s_%s' % (name, widget.consent_requirement.requirement_id))
            for widget in self.widgets
        ]

    def decompress(self, value):
        """Decompress a list of values for the widget.

        This is required by the parent class, and is responsible for taking the
        provided data and returning a list of values that can be used for the
        sub-widgets.

        Args:
            value (list):
                The list of values (or ``None``) to normalize and return.

        Returns:
            list of unicode:
            The resulting list of values. This may be empty.
        """
        return value or []


class ConsentRequirementField(forms.ChoiceField):
    """A form field for displaying and gathering consent information.

    This presents the consent requirement in an easy-to-digest form, and
    makes it easy for users to choose to allow or block the thing requiring
    consent.

    The cleaned result from this field is a
    :py:class:`~djblets.privacy.consent.base.ConsentData` instance, which
    can be recorded directly in the tracker.
    """

    ALLOW = 'allow'
    BLOCK = 'block'

    CHOICES = (
        (ALLOW, _('Allow')),
        (BLOCK, _('Block')),
    )

    widget = ConsentRequirementWidget

    default_error_messages = {
        'required': _('You must choose Allow or Block to continue.'),
    }

    def __init__(self, consent_requirement, user=None, consent_source=None,
                 extra_consent_data=None, **kwargs):
        """Initialize the field.

        Args:
            consent_requirement (djblets.privacy.consent.base.
                                 ConsentRequirement):
                The consent requirement information.

            user (django.contrib.auth.models.User, optional):
                The user viewing the form. If provided, the default value
                for the field will be based on the choice already made by
                the user, if any.

            consent_source (unicode, optional):
                The source to record in the consent audit trail for the
                consent choice saved in this field.

            extra_consent_data (dict, optional):
                Extra information to record in the consent audit trail for
                the consent choice saved in this field.

            **kwargs (dict):
                Additional keyword arguments to pass to the parent class.
        """
        super(ConsentRequirementField, self).__init__(
            choices=self.CHOICES,
            label='',
            required=True,
            **kwargs)

        self.consent_requirement = consent_requirement
        self.consent_source = consent_source
        self.extra_consent_data = extra_consent_data
        self.widget.consent_requirement = consent_requirement

        if user and user.is_authenticated:
            self.set_initial_from_user(user)

    def set_initial_from_user(self, user):
        """Set the initial state of the field based on a user's prior consent.

        This is called automatically if passing a user to the constructor.
        Otherwise, it should be called manually when setting up a form.

        Args:
            user (django.contrib.auth.models.User):
                The user viewing the form.
        """
        assert user
        assert user.is_authenticated

        self.initial = self.consent_requirement.get_consent(user)

    def prepare_value(self, value):
        """Prepare a value for use in the field.

        This will convert a :py:class:`~djblets.privacy.consent.base.Consent`
        value to a value suitable for use in the field.

        Args:
            value (djblets.privacy.consent.base.Consent):
                The value to convert.

        Returns:
            unicode:
            A valid value for use in the field.
        """
        return {
            Consent.GRANTED: self.ALLOW,
            Consent.DENIED: self.BLOCK,
        }.get(value)

    def clean(self, value):
        """Clean and return a value from submitted form data.

        Args:
            value (unicode):
                A value submitted by the client.

        Returns:
            djblets.privacy.consent.base.ConsentData:
            The cleaned consent data value, or ``None`` if a suitable value was
            not provided.
        """
        value = super(ConsentRequirementField, self).clean(value)

        # Validators should have already caught this.
        assert value in (self.ALLOW, self.BLOCK)

        return self.consent_requirement.build_consent_data(
            source=self.consent_source,
            extra_data=self.extra_consent_data,
            granted=(value == self.ALLOW))


class MultiConsentRequirementsField(forms.MultiValueField):
    """A form field for displaying and gathering mulltiple consent information.

    This provides a :py:class:`ConsentRequirementField` for each consent
    requirement provided to the field (or all registered ones if an explicit
    list was not provided). It's handy for forms that offer all consent choices
    to the user.

    The cleaned result from this field is a list of
    :py:class:`~djblets.privacy.consent.base.ConsentData` instances, which
    can be recorded directly in the tracker.
    """

    default_error_messages = {
        'required': _('You must choose Allow or Block for all options to '
                      'continue.'),
    }

    widget = MultiConsentRequirementsWidget

    def __init__(self, consent_requirements=None, user=None,
                 consent_source=None, extra_consent_data=None,
                 *args, **kwargs):
        """Initialize the field.

        Args:
            consent_requirements (list of djblets.privacy.consent.base.
                                  ConsentRequirement, optional):
                The list of consent requirements to display. If not provided,
                all registered consent requirements will be used.

            user (django.contrib.auth.models.User, optional):
                The user viewing the form. If provided, the default options
                for each field will be based on the choices already made by
                the user.

            consent_source (unicode, optional):
                The source to record in the consent audit trail for anything
                saved in this field.

            extra_consent_data (dict, optional):
                Extra information to record in the consent audit trail for
                anything saved in this field.

            *args (tuple):
                Additional positional arguments to pass to the parent class.

            **kwargs (dict):
                Additional keyword arguments to pass to the parent class.
        """
        self.consent_requirements = (consent_requirements or
                                     list(get_consent_requirements_registry()))

        super(MultiConsentRequirementsField, self).__init__(
            fields=[
                ConsentRequirementField(consent_requirement,
                                        consent_source=consent_source,
                                        extra_consent_data=extra_consent_data)
                for consent_requirement in self.consent_requirements
            ],
            widget=self.widget(self.consent_requirements),
            *args, **kwargs)

        if user and user.is_authenticated:
            self.set_initial_from_user(user)

    def set_initial_from_user(self, user):
        """Set the initial state of the field based on a user's prior consent.

        This is called automatically if passing a user to the constructor.
        Otherwise, it should be called manually when setting up a form.

        Args:
            user (django.contrib.auth.models.User):
                The user viewing the form.
        """
        assert user
        assert user.is_authenticated

        self.initial = [
            field.consent_requirement.get_consent(user)
            for field in self.fields
        ]

    def prepare_value(self, value):
        """Prepare a value for use in the field.

        This will convert a list of
        :py:class:`~djblets.privacy.consent.base.Consent` values given in the
        order of the field's list of requirements to values suitable for use
        in the field.

        Args:
            value (list of list djblets.privacy.consent.base.Consent):
                The list of values to convert.

        Returns:
            list of unicode:
            A list of values suitable for use in the field.
        """
        return [
            field.prepare_value(value[i])
            for i, field in enumerate(self.fields)
        ]

    def clean(self, value):
        """Clean and return values from submitted form data.

        Args:
            value (list of unicode):
                A list of values submitted by the client.

        Returns:
            list of djblets.privacy.consent.base.ConsentData:
            The list of cleaned consent data values.
        """
        consent_data_list = \
            super(MultiConsentRequirementsField, self).clean(value)

        # Set the timestamp to the same value on each entry.
        now = timezone.now()

        for consent_data in consent_data_list:
            consent_data.timestamp = now

        return consent_data_list

    def compress(self, data_list):
        """Compress cleaned values for the field.

        This is required by the parent class, and is responsible for taking a
        list of cleaned values and returning something that can be validated
        and returned. This implementation returns the data as-is.

        Args:
            data_list (list of djblets.privacy.consent.base.ConsentData):
                The list of cleaned data.

        Returns:
            list of djblets.privacy.consent.base.ConsentData:
            The list of data.
        """
        return data_list


class ConsentFormMixin(object):
    """A mixin for forms that present registered consent requirements.

    This can be mixed into a form to provide consent field initialization and
    saving.
    """

    #: The name of the consent field.
    consent_field_name = 'consent'

    def __init__(self, *args, **kwargs):
        """Initialize the form.

        Args:
            *args (tuple):
                Positional arguments passed to the form.

            **kwargs (dict):
                Keyword arguments passed to the form.
        """
        super(ConsentFormMixin, self).__init__(*args, **kwargs)

        self.fields[self.consent_field_name] = \
            MultiConsentRequirementsField(
                label='',
                user=self.get_consent_user(),
                consent_source=self.get_consent_source(),
                extra_consent_data=self.get_extra_consent_data())

    def get_consent_user(self):
        """Return the user deciding on consent.

        This must be implemented by subclasses.

        Returns:
            django.contrib.auth.models.User:
            The user deciding on consent.
        """
        raise NotImplementedError

    def get_consent_source(self):
        """Return a source to record in the consent audit trail.

        This must be implemented by subclasses.

        Returns:
            unicode:
            The source to record for each consent entry.
        """
        raise NotImplementedError

    def get_extra_consent_data(self):
        """Return extra data to record in the consent audit trail.

        By default, this just returns an empty dictionary.

        Returns:
            dict:
            Extra data to record for each consent entry.
        """
        return {}

    def save_consent(self, user):
        """Save the consent information recorded in the form.

        Args:
            user (django.contrib.auth.models.User):
                The user who made the consent decisions.
        """
        assert self.is_valid()
        assert user
        assert user.is_authenticated

        get_consent_tracker().record_consent_data_list(
            user,
            self.cleaned_data[self.consent_field_name])


class ConsentConfigPageFormMixin(ConsentFormMixin):
    """A mixin for config forms that present registered consent requirements.

    This can be mixed into a config form to provide consent field
    initialization and saving. It would be used instead of
    :py:class:`ConsentFormMixin`.
    """

    form_id = 'privacy_consent'
    form_title = _('Privacy Consent')

    def save(self):
        """Save the form.

        This will save the consent information from the field.
        """
        self.save_consent(self.request.user)

        messages.add_message(
            self.request, messages.INFO,
            _('Your choices have been saved. You can make changes to these '
              'at any time.'))

    def get_consent_user(self):
        """Return the user deciding on consent.

        Returns:
            django.contrib.auth.models.User:
            The user deciding on consent.
        """
        return self.request.user

    def get_consent_source(self):
        """Return a source to record in the consent audit trail.

        By default, this returns the absolute URL for the page.

        Returns:
            unicode:
            The source to record for each consent entry.
        """
        return self.request.build_absolute_uri()



"""Utility functions for looking up static media URLs."""

from django.templatetags.static import static
from django.utils.functional import lazy


#: Lazily look up a static media URL.
#:
#: This is used to reference static media URLs in a class attribute, in a
#: global variable, or some other definition that would be evaluated upon
#: loading the module. Calling :py:func:`~django.templatetags.static`
#: will immediately perform a lookup, which may be too early in the startup
#: sequence of the consuming application. By lazily referencing the URL,
#: the lookup can be performed later when needed, and not on module load.
#:
#: Args:
#:     path (unicode):
#:         The static media path.
#:
#: Returns:
#:     django.utils.functional.__proxy__:
#:     A proxy that evaluates to the static media URL (as a unicode string).
static_lazy = lazy(static, str)



"""Web API signals."""

from django.core.signals import Signal


#: A signal indicating a WebAPI token has been created.
#:
#: Args:
#:     instance (djblets.webapi.models.WebAPIToken):
#:         The created instance.
#:
#:     auto_generated (bool):
#:         Whether or not the token was automatically generated.
webapi_token_created = Signal()


#: A signal indicating an attempt to authenticate with an expired WebAPI token.
#:
#: Version Added:
#:     3.0
#:
#: Args:
#:     instance (djblets.webapi.models.WebAPIToken):
#:         The expired instance.
webapi_token_expired = Signal()


#: A signal indicating a WebAPI token has been updated.
#:
#: Args:
#:     instance (djblets.webapi.models.WebAPIToken):
#:         The updated instance.
webapi_token_updated = Signal()



import json

from django.core.management.base import BaseCommand

from djblets.siteconfig.models import SiteConfiguration


class Command(BaseCommand):
    """Lists the site configuration."""

    def handle(self, *args, **options):
        siteconfig = SiteConfiguration.objects.get_current()

        self.stdout.write(json.dumps(siteconfig.settings, indent=2))



"""Privacy-related template tags."""

from django import template
from django.utils.html import format_html

from djblets.privacy.pii import build_pii_safe_page_url_for_request


register = template.Library()


@register.simple_tag(takes_context=True)
def pii_safe_page_url(context):
    """Inject the current page URL with personal information redacted.

    This makes use of :py:func:`djblets.privacy.pii.build_pii_safe_page_url`
    to inject a version of the current page URL with usernames or e-mail
    addresses redacted.

    Args:
        context (django.templates.RequestContext):
            The context for the page, containing a ``request`` variable.

    Returns:
        unicode:
        The safe URL to inject into the page.

    Example:
        .. code-block:: html+django

           ga('set', 'location', '{% pii_safe_page_url %}');
    """
    return format_html(
        '{}',
        build_pii_safe_page_url_for_request(context['request']))



"""Middleware for extensions."""

import threading

from django.conf import settings
from django.utils.deprecation import MiddlewareMixin

from djblets.extensions.manager import get_extension_managers


class ExtensionsMiddleware(MiddlewareMixin):
    """Middleware to manage extension lifecycles and data."""

    def __init__(self, *args, **kwargs):
        """Initialize the middleware.

        Args:
            *args (tuple):
                Positional arguments to pass through to the superclass.

            **kwargs (dict):
                Keyword arguments to pass through to the superclass.
        """
        super(ExtensionsMiddleware, self).__init__(*args, **kwargs)

        self.do_expiration_checks = not getattr(settings, 'RUNNING_TEST',
                                                False)
        self._lock = threading.Lock()

    def process_request(self, request):
        if self.do_expiration_checks:
            self._check_expired()

    def process_view(self, request, view, args, kwargs):
        request._djblets_extensions_kwargs = kwargs

    def _check_expired(self):
        """Checks each ExtensionManager for expired extension state.

        When the list of extensions on an ExtensionManager changes, or when
        the configuration of an extension changes, any other threads/processes
        holding onto extensions and configuration will go stale. This function
        will check each of those to see if they need to re-load their
        state.

        This is meant to be called before every HTTP request.
        """
        for extension_manager in get_extension_managers():
            # We're going to check the expiration, and then only lock if it's
            # expired. Following that, we'll check again.
            #
            # We do this in order to prevent locking unnecessarily, which could
            # impact performance or cause a problem if a thread is stuck.
            #
            # We're checking the expiration twice to prevent every blocked
            # thread from making its own attempt to reload the extensions once
            # the first thread holding the lock finishes the reload.
            if extension_manager.is_expired():
                with self._lock:
                    # Check again, since another thread may have already
                    # reloaded.
                    if extension_manager.is_expired():
                        extension_manager.load(full_reload=True)


class ExtensionsMiddlewareRunner(MiddlewareMixin):
    """Middleware to execute middleware from extensions.

    The process_*() methods iterate over all extensions' middleware, calling
    the given method if it exists. The semantics of how Django executes each
    method are preserved.

    This middleware should be loaded after the main extension middleware
    (djblets.extensions.middleware.ExtensionsMiddleware). It's probably
    a good idea to have it be at the very end so that everything else in the
    core that needs to be initialized is done before any extension's
    middleware is run.
    """

    def process_view(self, request, view_func, view_args, view_kwargs):
        """Process a view through extension middleware.

        Args:
            request (django.http.HttpRequest):
                The request object.

            view_func (callable):
                The view callable.

            view_args (list):
                Positional arguments to pass to the view.

            view_kwargs (dict):
                Keyword arguments to pass to the view.

        Returns:
            django.http.HttpResponse or None:
            Either a response object (in which case other middleware will not
            be run), or None.
        """
        for cls in self._middleware_classes:
            middleware = cls(self.get_response)

            if hasattr(middleware, 'process_view'):
                result = middleware.process_view(request, view_func, view_args,
                                                 view_kwargs)

                if result:
                    return result

        return None

    def process_template_response(self, request, response):
        """Process a template response through extension middleware.

        Args:
            request (django.http.HttpRequest):
                The request object.

            response (django.http.HttpResponse):
                The response from the view.

        Returns:
            django.template.response.TemplateResponse:
            A new template response to execute.
        """
        for cls in self._middleware_classes:
            middleware = cls(self.get_response)

            if hasattr(middleware, 'process_template_response'):
                response = middleware.process_template_response(
                    request, response)

        return response

    def process_exception(self, request, exception):
        """Process an exception through extension middleware.

        Args:
            request (django.http.HttpRequest):
                The request object.

            exception (Exception):
                The exception to process.

        Returns:
            django.http.HttpResponse or None:
            Either a response object (in which case other middleware will not
            be run), or None.
        """
        for cls in self._middleware_classes:
            middleware = cls(self.get_response)

            if hasattr(middleware, 'process_exception'):
                result = middleware.process_exception(request, exception)

                if result:
                    return result

        return None

    def __call__(self, request):
        """Run extension middleware.

        Args:
            request (django.http.HttpRequest):
                The HTTP request object.

        Returns:
            django.http.HttpResponse:
            The HTTP response.
        """
        middleware = list(self._middleware_classes)
        middleware.reverse()

        def _get_response_iter(request, middleware=middleware):
            try:
                next_method = middleware.pop(0)(_get_response_iter)
            except IndexError:
                next_method = self.get_response

            return next_method(request)

        return _get_response_iter(request)

    @property
    def _middleware_classes(self):
        """All extension-provided middleware.

        Yields:
            callable:
            Extension-provided middleware (either class or function).
        """
        for mgr in get_extension_managers():
            for middleware in mgr.middleware_classes:
                yield middleware



from django.forms import Form, ValidationError

from djblets.conditions.choices import (BaseConditionChoice,
                                        BaseConditionIntegerChoice,
                                        BaseConditionStringChoice,
                                        ConditionChoices)
from djblets.conditions.conditions import ConditionSet
from djblets.forms.fields import ConditionsField
from djblets.testing.testcases import TestCase


class ConditionsFieldTests(TestCase):
    """Unit tests for djblets.forms.fields.ConditionsField."""

    def test_init_with_choices_instance(self):
        """Testing ConditionsField initialization with choices instance"""
        choices = ConditionChoices([BaseConditionStringChoice])
        field = ConditionsField(choices=choices)

        self.assertIs(field.choices, choices)

    def test_init_with_choices_subclass(self):
        """Testing ConditionsField initialization with choices subclass"""
        class MyChoices(ConditionChoices):
            choice_classes = [BaseConditionStringChoice]

        field = ConditionsField(choices=MyChoices)

        self.assertIs(field.choices.__class__, MyChoices)

    def test_init_with_choice_kwargs(self):
        """Testing ConditionsField initialization with choice_kwargs"""
        choices = ConditionChoices([BaseConditionStringChoice])
        field = ConditionsField(choices=choices, choice_kwargs={'abc': 123})

        self.assertEqual(field.widget.choice_kwargs, field.choice_kwargs)

    def test_init_with_missing_operators(self):
        """Testing ConditionsField initialization with choices missing
        operators
        """
        class MyChoice(BaseConditionChoice):
            choice_id = 'my-choice'

        choices = ConditionChoices([MyChoice])
        msg = 'MyChoice must define a non-empty "operators" attribute.'

        with self.assertRaisesMessage(ValueError, msg):
            ConditionsField(choices=choices)

    def test_choice_kwargs_with_multiple_instances(self):
        """Testing ConditionsField.choice_kwargs works with multiple instances
        """
        # This tests for a bug that existed prior to Djblets 1.0.10, where
        # data set in choice_kwargs in one instance of a form would get shared
        # amongst other instances.
        class MyForm(Form):
            conditions = ConditionsField(
                choices=ConditionChoices([BaseConditionStringChoice]))

        form1 = MyForm()
        form2 = MyForm()

        form1.fields['conditions'].choice_kwargs['a'] = 1

        self.assertNotIn('a', form2.fields['conditions'].choice_kwargs)

    def test_prepare_value_with_condition_set(self):
        """Testing ConditionsField.prepare_value with ConditionSet"""
        choices = ConditionChoices([BaseConditionStringChoice])
        field = ConditionsField(choices=choices)

        self.assertEqual(
            field.prepare_value(ConditionSet()),
            {
                'mode': 'all',
                'conditions': [],
            })

    def test_prepare_value_with_serialized_data(self):
        """Testing ConditionsField.prepare_value with serialized data"""
        choices = ConditionChoices([BaseConditionStringChoice])
        field = ConditionsField(choices=choices)

        data = {
            'mode': 'all',
            'conditions': [],
        }

        self.assertEqual(field.prepare_value(data), data)

    def test_to_python(self):
        """Testing ConditionsField.to_python"""
        class MyChoice(BaseConditionStringChoice):
            choice_id = 'my-choice'

        choices = ConditionChoices([MyChoice])
        field = ConditionsField(choices=choices)

        condition_set = field.to_python({
            'mode': 'any',
            'conditions': [{
                'choice': 'my-choice',
                'op': 'is',
                'value': 'my-value',
            }]
        })

        self.assertEqual(condition_set.mode, ConditionSet.MODE_ANY)
        self.assertEqual(len(condition_set.conditions), 1)

        condition = condition_set.conditions[0]
        self.assertEqual(condition.choice.choice_id, 'my-choice')
        self.assertEqual(condition.operator.operator_id, 'is')
        self.assertEqual(condition.value, 'my-value')

    def test_to_python_with_choice_kwargs(self):
        """Testing ConditionsField.to_python with choice_kwargs set"""
        class MyChoice(BaseConditionStringChoice):
            choice_id = 'my-choice'

        choices = ConditionChoices([MyChoice])
        field = ConditionsField(choices=choices,
                                choice_kwargs={'abc': 123})

        condition_set = field.to_python({
            'mode': 'any',
            'conditions': [{
                'choice': 'my-choice',
                'op': 'is',
                'value': 'my-value',
            }]
        })

        self.assertEqual(condition_set.mode, ConditionSet.MODE_ANY)
        self.assertEqual(len(condition_set.conditions), 1)

        choice = condition_set.conditions[0].choice
        self.assertEqual(choice.choice_id, 'my-choice')
        self.assertEqual(choice.extra_state, {'abc': 123})

    def test_to_python_with_mode_error(self):
        """Testing ConditionsField.to_python with mode error"""
        choices = ConditionChoices()
        field = ConditionsField(choices=choices)

        with self.assertRaises(ValidationError) as cm:
            field.to_python({
                'mode': 'invalid',
                'conditions': [],
            })

        self.assertEqual(cm.exception.messages,
                         ['"invalid" is not a valid condition mode.'])
        self.assertEqual(cm.exception.code, 'invalid_mode')

    def test_to_python_with_choice_not_found_error(self):
        """Testing ConditionsField.to_python with choice not found error"""
        choices = ConditionChoices()
        field = ConditionsField(choices=choices)

        with self.assertRaises(ValidationError) as cm:
            field.to_python({
                'mode': 'any',
                'conditions': [{
                    'choice': 'invalid-choice',
                    'op': 'is',
                    'value': 'my-value',
                }],
            })

        self.assertEqual(cm.exception.messages,
                         ['There was an error with one of your conditions.'])
        self.assertEqual(cm.exception.code, 'condition_errors')
        self.assertEqual(
            field.widget.condition_errors,
            {
                0: 'No condition choice was found matching "invalid-choice".',
            })

    def test_to_python_with_operator_not_found_error(self):
        """Testing ConditionsField.to_python with operator not found error"""
        class MyChoice(BaseConditionStringChoice):
            choice_id = 'my-choice'

        choices = ConditionChoices([MyChoice])
        field = ConditionsField(choices=choices)

        with self.assertRaises(ValidationError) as cm:
            field.to_python({
                'mode': 'any',
                'conditions': [{
                    'choice': 'my-choice',
                    'op': 'invalid-op',
                    'value': 'my-value',
                }],
            })

        self.assertEqual(cm.exception.messages,
                         ['There was an error with one of your conditions.'])
        self.assertEqual(cm.exception.code, 'condition_errors')
        self.assertEqual(
            field.widget.condition_errors,
            {
                0: 'No operator was found matching "invalid-op".',
            })

    def test_to_python_with_invalid_value_error(self):
        """Testing ConditionsField.to_python with invalid value error"""
        class MyChoice(BaseConditionIntegerChoice):
            choice_id = 'my-choice'

        choices = ConditionChoices([MyChoice])
        field = ConditionsField(choices=choices)

        with self.assertRaises(ValidationError) as cm:
            field.to_python({
                'mode': 'any',
                'conditions': [{
                    'choice': 'my-choice',
                    'op': 'is',
                    'value': 'invalid-value',
                }],
            })

        self.assertEqual(cm.exception.messages,
                         ['There was an error with one of your conditions.'])
        self.assertEqual(cm.exception.code, 'condition_errors')
        self.assertEqual(
            field.widget.condition_errors,
            {
                0: 'Enter a whole number.',
            })

    def test_to_python_with_value_required_error(self):
        """Testing ConditionsField.to_python with value required error"""
        class MyChoice(BaseConditionStringChoice):
            choice_id = 'my-choice'

        choices = ConditionChoices([MyChoice])
        field = ConditionsField(choices=choices)

        with self.assertRaises(ValidationError) as cm:
            field.to_python({
                'mode': 'any',
                'conditions': [{
                    'choice': 'my-choice',
                    'op': 'is',
                }],
            })

        self.assertEqual(cm.exception.messages,
                         ['There was an error with one of your conditions.'])
        self.assertEqual(cm.exception.code, 'condition_errors')
        self.assertEqual(
            field.widget.condition_errors,
            {
                0: 'A value is required.',
            })



from django.contrib.auth.models import User
from django.core.exceptions import ValidationError
from django.db import models
from django.utils import timezone
from django.utils.translation import gettext_lazy as _

from djblets.db.fields import JSONField, ModificationTimestampField
from djblets.secrets.token_generators import token_generator_registry
from djblets.webapi.managers import WebAPITokenManager
from djblets.webapi.signals import webapi_token_updated


class BaseWebAPIToken(models.Model):
    """Base class for an access token used for authenticating with the API.

    Each token can be used to authenticate the token's owner with the API,
    without requiring a username or password to be provided. Tokens can
    be revoked, and new tokens added.

    Tokens can store policy information, which will later be used for
    restricting access to the API.
    """

    user = models.ForeignKey(
        User,
        related_name='webapi_tokens',
        on_delete=models.CASCADE,
        help_text=_('The user that owns the token.'))

    token = models.CharField(
        max_length=255,
        unique=True,
        help_text=_('The access token.'))
    token_generator_id = models.CharField(
        max_length=255,
        help_text=_('The token generator that generated the token.'))
    time_added = models.DateTimeField(
        default=timezone.now,
        help_text=_('The date and time when the token was first added '
                    'to the database.'))
    last_updated = ModificationTimestampField(
        default=timezone.now,
        help_text=_('The date and time when the token was last updated.'))
    last_used = models.DateTimeField(
        null=True,
        blank=True,
        help_text=_('The date and time when the token was last used '
                    'for authentication.'))
    expires = models.DateTimeField(
        null=True,
        blank=True,
        help_text=_('An optional field for the date and time that the token '
                    'will expire. The token will be invalid and unusable '
                    'for authentication after this point.'))

    valid = models.BooleanField(
        default=True,
        help_text=_('Whether the token is currently valid.'))
    invalid_date = models.DateTimeField(
        null=True,
        blank=True,
        help_text=_('The date and time at which the token became invalid.'))
    invalid_reason = models.TextField(
        blank=True,
        help_text=_('A message indicating why the token is no longer valid.'))

    note = models.TextField(
        blank=True,
        help_text=_('A message describing the token.'))
    policy = JSONField(
        null=True,
        help_text=_('The policy document describing what this token can '
                    'access in the API. If empty, this provides full access.'))

    extra_data = JSONField(null=True)

    objects = WebAPITokenManager()

    def is_accessible_by(self, user):
        return user.is_superuser or self.user == user

    def is_mutable_by(self, user):
        return user.is_superuser or self.user == user

    def is_deletable_by(self, user):
        return user.is_superuser or self.user == user

    def is_deprecated(self) -> bool:
        """Return whether the token is deprecated or not.

        This will be ``True`` if the token was generated by a deprecated
        token generator.

        Version Added:
            3.1

        Returns:
            bool:
            Whether the token is deprecated.
        """
        generators = token_generator_registry.get_deprecated_token_generators()

        return self.token_generator_id in generators

    def is_expired(self) -> bool:
        """Return whether the token is expired or not.

        This will be ``False`` if there is no expiration date set.

        Version Added:
            3.0

        Returns:
            bool:
            Whether the token is expired.
        """
        return self.expires is not None and timezone.now() >= self.expires

    def __str__(self):
        return 'Web API token for %s' % self.user

    def save(self, *args, **kwargs):
        """Save the token.

        If the token is being updated, the
        :py:data:`~djblets.webapi.signals.webapi_token_updated` signal will be
        emitted.

        Args:
            *args (tuple):
                Positional arguments to pass to the superclass.

            **kwargs (dict):
                Keyword arguments to pass to the superclass.
        """
        is_new = self.pk is None

        super().save(*args, **kwargs)

        if not is_new:
            webapi_token_updated.send(instance=self, sender=type(self))

    @classmethod
    def get_root_resource(cls):
        raise NotImplementedError

    @classmethod
    def validate_policy(cls, policy):
        """Validate an API policy.

        This will check to ensure that the policy is in a suitable format
        and contains the information required in a format that can be parsed.

        If a failure is found, a ValidationError will be raised describing
        the error and where it was found.
        """
        if not isinstance(policy, dict):
            raise ValidationError(_('The policy must be a JSON object.'))

        if not policy:
            # Empty policies are equivalent to allowing full access. If this
            # is empty, we can stop here.
            return

        if 'resources' not in policy:
            raise ValidationError(
                _('The policy is missing a "resources" section.'))

        resources_section = policy['resources']

        if not isinstance(resources_section, dict):
            raise ValidationError(
                _('The policy\'s "resources" section must be a JSON object.'))

        if not resources_section:
            raise ValidationError(
                _('The policy\'s "resources" section must not be empty.'))

        if '*' in resources_section:
            cls._validate_policy_section(resources_section, '*',
                                         'resources.*')

        resource_policies = [
            (section_name, section)
            for section_name, section in resources_section.items()
            if section_name != '*'
        ]

        if resource_policies:
            valid_policy_ids = \
                cls._get_valid_policy_ids(cls.get_root_resource())

            for policy_id, section in resource_policies:
                if policy_id not in valid_policy_ids:
                    raise ValidationError(
                        _('"%s" is not a valid resource policy ID.')
                        % policy_id)

                for subsection_name, subsection in section.items():
                    if not isinstance(subsection_name, str):
                        raise ValidationError(
                            _('%s must be a string in "resources.%s"')
                            % (subsection_name, policy_id))

                    cls._validate_policy_section(
                        section, subsection_name,
                        'resources.%s.%s' % (policy_id, subsection_name))

    @classmethod
    def _validate_policy_section(cls, parent_section, section_name,
                                 full_section_name):
        section = parent_section[section_name]

        if not isinstance(section, dict):
            raise ValidationError(
                _('The "%s" section must be a JSON object.')
                % full_section_name)

        if 'allow' not in section and 'block' not in section:
            raise ValidationError(
                _('The "%s" section must have "allow" and/or "block" '
                  'rules.')
                % full_section_name)

        if 'allow' in section and not isinstance(section['allow'], list):
            raise ValidationError(
                _('The "%s" section\'s "allow" rule must be a list.')
                % full_section_name)

        if 'block' in section and not isinstance(section['block'], list):
            raise ValidationError(
                _('The "%s" section\'s "block" rule must be a list.')
                % full_section_name)

    @classmethod
    def _get_valid_policy_ids(cls, resource, result=None):
        if result is None:
            result = set()

        if hasattr(resource, 'policy_id'):
            result.add(resource.policy_id)

        for child_resource in resource.list_child_resources:
            cls._get_valid_policy_ids(child_resource, result)

        for child_resource in resource.item_child_resources:
            cls._get_valid_policy_ids(child_resource, result)

        return result

    class Meta:
        abstract = True
        verbose_name = _('Web API token')
        verbose_name_plural = _('Web API tokens')



"""Field for managing modification timestamps for a model."""

from django.db import models
from django.utils import timezone


class ModificationState(object):
    """Modification state for a tracked value.

    Attributes:
        first_set (bool):
            Whether or not the next :py:func:`setattr` will be the first.

            When this is ``True``, we do not consider the next
            :py:func:`setattr` to be a modification since the field will be
            set during model initialization.

        modified (bool):
            Whether or not the value has been manually modified.
    """

    def __init__(self):
        """Initialize the state."""
        self.modified = False
        self.first_set = False

    @staticmethod
    def get_attr_name(field_name):
        """Return the attribute name for the state for the named field.

        Args:
            field_name (unicode):
                The attribute name of the
                :py:class:`ModificationTimestampField`.

        Returns:
            unicode:
            The name of the modification state attribute.
        """
        return '_%s_state' % field_name


class ModificationTrackedValue(object):
    """A descriptor for tracking the modification of a value.

    Attributes:
        field_name (unicode):
            The attribute name of the field on the model.

        state_name (unicode):
            The attribute name of the state on the model.
    """

    def __init__(self, field_name, state_name):
        """Initialize the descriptor.

        Args:
            field_name (unicode):
                The name of the field attribute.

            state_name (unicode):
                The name of the state attribute.
        """
        self.field_name = field_name
        self.state_name = state_name

    def __get__(self, instance, objtype=None):
        """Return the value.

        Args:
            instance (django.db.models.Model):
                The model instance.

            objtype (type, optional):
                The model class.

        Returns:
            datetime.datetime:
            The value.
        """
        return instance.__dict__[self.field_name]

    def __set__(self, instance, value):
        """Set the value.

        This tracks modifications and updates the state if this is not the
        first call to this method (which occurs during model initialization).

        Args:
            instance (django.db.models.Model):
                The instance to set the the value on.

            value (datetime.datetime):
                The value to set.state.first_set:
        """
        state = getattr(instance, self.state_name)

        if state.first_set:
            state.first_set = False
        else:
            state.modified = True

        instance.__dict__[self.field_name] = value


class ModificationTimestampField(models.DateTimeField):
    """A timestamp field that only updates existing objects or when None.

    This is a subclass of :py:class:`~django.db.models.DateTimeField` that only
    auto-updates the timestamp when updating an existing object or when the
    value of the field is None. It's similar to using ``auto_now=True``, but
    custom timestamp values on new instances will not be replaced.
    """

    def __init__(self, *args, **kwargs):
        """Initialize the field.

        Args:
            *args (tuple):
                Positional arguments for the field.

            **kwargs (dict):
                Additional keyword arguments for the field.
        """
        kwargs.update({
            'editable': False,
            'blank': True,
        })
        super(ModificationTimestampField, self).__init__(*args, **kwargs)

    def contribute_to_class(self, cls, name):
        """Contribute the field attributes to the class.

        Args:
            cls (type):
                The model class to contribute the field to.

            name (unicode):
                The name of the field.
        """
        super(ModificationTimestampField, self).contribute_to_class(cls, name)

        state_name = ModificationState.get_attr_name(self.name)
        setattr(cls, self.name,
                ModificationTrackedValue(self.name, state_name))
        setattr(cls, state_name, ModificationState())

    def pre_save(self, model_instance, add):
        """Return the value of the field just before saving.

        Args:
            model_instance (django.db.models.Model):
                The model instance being saved.

            add (bool):
                Whether this is being saved to the database for the first time.

        Returns:
            datetime.datetime:
            The date/time value being saved.
        """
        state = getattr(model_instance,
                        ModificationState.get_attr_name(self.name))
        existing_value = getattr(model_instance, self.attname)

        if ((not add and not state.modified) or
            (add and existing_value is None)):
            value = timezone.now()
            setattr(model_instance, self.attname, value)
        elif ((not add and state.modified) or
              (add and existing_value is not None)):
            value = super(ModificationTimestampField, self).pre_save(
                model_instance, add)

        state.modified = False
        return value



import re

from django import forms
from django.contrib.sites.models import Site
from django.db.models.query import QuerySet
from django.http import QueryDict

from djblets.conditions.errors import InvalidConditionValueError
from djblets.conditions.values import (ConditionValueBooleanField,
                                       ConditionValueFormField,
                                       ConditionValueIntegerField,
                                       ConditionValueModelField,
                                       ConditionValueMultipleModelField,
                                       ConditionValueRegexField)
from djblets.testing.testcases import TestCase


class ConditionValueFormFieldTests(TestCase):
    """Unit tests for djblets.conditions.values.ConditionValueFormField."""

    def test_serialize_value(self):
        """Testing ConditionValueFormField.serialize_value"""
        # ModelChoiceField uses prepare_value(), needed by serialize_value(),
        # so it's a good test.
        field = ConditionValueFormField(
            forms.ModelChoiceField(queryset=Site.objects.all()))

        site = Site.objects.get_current()
        self.assertEqual(field.serialize_value(site), site.pk)

    def test_deserialize_value(self):
        """Testing ConditionValueFormField.deserialize_value"""
        field = ConditionValueFormField(
            forms.ModelChoiceField(queryset=Site.objects.all()))

        site = Site.objects.get_current()
        self.assertEqual(field.deserialize_value(site.pk), site)

    def test_get_from_form_data(self):
        """Testing ConditionValueFormField.get_from_form_data"""
        field = ConditionValueFormField(forms.CharField())

        data = QueryDict('test-field=abc123')
        files = QueryDict('')

        self.assertEqual(field.get_from_form_data(data, files, 'test-field'),
                         'abc123')

    def test_render_html(self):
        """Testing ConditionValueFormField.render_html"""
        field = ConditionValueFormField(forms.CharField())

        self.assertHTMLEqual(field.render_html(),
                             '<input name="XXX" type="text" />')

    def test_field_with_callable(self):
        """Testing ConditionValueFormField.field with callable field"""
        field = ConditionValueFormField(lambda: forms.CharField())

        self.assertTrue(isinstance(field.field, forms.CharField))


class ConditionValueBooleanFieldTests(TestCase):
    """Unit tests for djblets.conditions.values.ConditionValueBooleanField."""

    def setUp(self):
        super(ConditionValueBooleanFieldTests, self).setUp()

        self.value_field = ConditionValueBooleanField()

    def test_deserialize_value_with_true(self):
        """Testing ConditionValueBooleanField.deserialize_value with True"""
        self.assertTrue(self.value_field.deserialize_value(True))
        self.assertTrue(self.value_field.deserialize_value('true'))
        self.assertTrue(self.value_field.deserialize_value('True'))
        self.assertTrue(self.value_field.deserialize_value('1'))

    def test_deserialize_value_with_false(self):
        """Testing ConditionValueBooleanField.deserialize_value with False"""
        self.assertFalse(self.value_field.deserialize_value(False))
        self.assertFalse(self.value_field.deserialize_value('false'))
        self.assertFalse(self.value_field.deserialize_value('False'))
        self.assertFalse(self.value_field.deserialize_value('0'))

    def test_prepare_value_for_widget(self):
        """Testing ConditionValueBooleanField.prepare_value_for_widget"""
        self.assertEqual(self.value_field.prepare_value_for_widget(True),
                         'True')
        self.assertEqual(self.value_field.prepare_value_for_widget(False),
                         'False')


class ConditionValueIntegerFieldTests(TestCase):
    """Unit tests for djblets.conditions.values.ConditionValueIntegerField."""

    def setUp(self):
        super(ConditionValueIntegerFieldTests, self).setUp()

        self.value_field = ConditionValueIntegerField()

    def test_deserialize_value(self):
        """Testing ConditionValueIntegerField.deserialize_value"""
        self.assertEqual(self.value_field.deserialize_value(100), 100)
        self.assertEqual(self.value_field.deserialize_value('100'), 100)


class ConditionValueModelFieldTests(TestCase):
    """Unit tests for ConditionValueModelField."""

    def test_init_with_queryset(self):
        """Testing ConditionValueModelField initialization with QuerySet"""
        field = ConditionValueModelField(queryset=Site.objects.all())

        self.assertTrue(isinstance(field.field.queryset, QuerySet))

    def test_init_with_callable(self):
        """Testing ConditionValueModelField initialization with callable"""
        field = ConditionValueModelField(queryset=lambda: Site.objects.all())

        self.assertTrue(isinstance(field.field.queryset, QuerySet))

    def test_serialize_value(self):
        """Testing ConditionValueModelField.serialize_value"""
        field = ConditionValueModelField(queryset=Site.objects.all())

        site = Site.objects.get_current()
        self.assertEqual(field.serialize_value(site), site.pk)

    def test_deserialize_value(self):
        """Testing ConditionValueModelField.deserialize_value"""
        field = ConditionValueModelField(queryset=Site.objects.all())

        site = Site.objects.get_current()
        self.assertEqual(field.deserialize_value(site.pk), site)


class ConditionValueMultipleModelFieldTests(TestCase):
    """Unit tests for ConditionValueMultipleModelField."""

    def test_init_with_queryset(self):
        """Testing ConditionValueMultipleModelField initialization with
        QuerySet
        """
        field = ConditionValueMultipleModelField(queryset=Site.objects.all())

        self.assertTrue(isinstance(field.field.queryset, QuerySet))

    def test_init_with_callable(self):
        """Testing ConditionValueMultipleModelField initialization with
        callable
        """
        field = ConditionValueMultipleModelField(
            queryset=lambda: Site.objects.all())

        self.assertTrue(isinstance(field.field.queryset, QuerySet))

    def test_serialize_value(self):
        """Testing ConditionValueMultipleModelField.serialize_value"""
        field = ConditionValueMultipleModelField(queryset=Site.objects.all())

        site = Site.objects.get_current()
        self.assertEqual(field.serialize_value([site]), [site.pk])

    def test_deserialize_value(self):
        """Testing ConditionValueMultipleModelField.deserialize_value"""
        field = ConditionValueMultipleModelField(queryset=Site.objects.all())

        site = Site.objects.get_current()
        result = field.deserialize_value([site.pk])

        self.assertEqual(len(result), 1)
        self.assertEqual(result[0], site)


class ConditionValueRegexFieldTests(TestCase):
    """Unit tests for djblets.conditions.values.ConditionValueRegexField."""

    def setUp(self):
        super(ConditionValueRegexFieldTests, self).setUp()

        self.value_field = ConditionValueRegexField()

    def test_deserialize_value(self):
        """Testing ConditionValueRegexField.deserialize_value"""
        regex_obj = self.value_field.deserialize_value('ab[^c]d+e?')

        # Weird, right? re.RegexObject, the documented result of a
        # re.compile(), does not exist. Instead, it's a _sre.SRE_Pattern,
        # but that's basically an internal detail we don't want to rely upon.
        # Internally, the re module actually does this type() mess to get
        # the type to use, so that's what we're doing!
        regex_type = type(re.compile(''))

        self.assertTrue(isinstance(regex_obj, regex_type))
        self.assertEqual(regex_obj.pattern, re.compile('ab[^c]d+e?').pattern)

    def test_deserialize_value_with_bad_pattern(self):
        """Testing ConditionValueRegexField.deserialize_value with bad
        pattern
        """
        with self.assertRaises(InvalidConditionValueError):
            self.value_field.deserialize_value('*')

    def test_serialize_value(self):
        """Testing ConditionValueRegexField.serialize_value"""
        regex_obj = re.compile('ab[^c]d+e?')

        self.assertEqual(
            self.value_field.serialize_value(regex_obj),
            'ab[^c]d+e?')



"""Djblets registries.

Registries are collections that keep track of unique objects.

For information on writing registries, see
:ref:`the guide on writing registries <writing-registries>`.
"""

import logging
from typing import (Dict, Generic, Iterable, Iterator, List, Optional,
                    Sequence, Set, Type, TypeVar)

from django.utils.translation import gettext_lazy as _
from pkg_resources import EntryPoint, iter_entry_points

from djblets.registries.errors import (AlreadyRegisteredError,
                                       ItemLookupError,
                                       RegistrationError)
from djblets.registries.signals import registry_populating


ALREADY_REGISTERED = 'already_registered'
ATTRIBUTE_REGISTERED = 'attribute_registered'
INVALID_ATTRIBUTE = 'invalid_attribute'
MISSING_ATTRIBUTE = 'missing_attribute'
UNREGISTER = 'unregister'
NOT_REGISTERED = 'not_registered'
LOAD_ENTRY_POINT = 'load_entry_point'


#: Default error messages for registries.
DEFAULT_ERRORS = {
    ALREADY_REGISTERED: _(
        'Could not register %(item)s: it is already registered.'
    ),
    ATTRIBUTE_REGISTERED: _(
        'Could not register %(item)s: another item (%(duplicate)s) is already '
        'registered with %(attr_name)s = %(attr_value)s.'
    ),
    INVALID_ATTRIBUTE: _(
        '"%(attr_name)s" is not a registered lookup attribute.'
    ),
    LOAD_ENTRY_POINT: _(
        'Could not load entry point %(entry_point)s: %(error)s.',
    ),
    MISSING_ATTRIBUTE: _(
        'Could not register %(item)s: it does not have a "%(attr_name)s" '
        'attribute.'
    ),
    UNREGISTER: _(
        'Could not unregister %(item)s: it is not registered.'
    ),
    NOT_REGISTERED: _(
        'No item registered with %(attr_name)s = %(attr_value)s.'
    ),
}


#: A generic type for items stored in a registry.
#:
#: This can be used for subclasses of :py:class:`Registry`, mixins, or other
#: utility code that need to stay generic. In normal usage, an explicit type
#: will be provided when subclassing instead.
#:
#: Version Added:
#:     3.1
RegistryItemType = TypeVar('RegistryItemType')


class Registry(Generic[RegistryItemType]):
    """An item registry.

    Item registries hold a set of objects that can be looked up by attributes.
    Each item is guaranteed to be unique and not share these attributes with
    any other item in the registry.

    Registries default to holding arbitrary objects. To limit objects to a
    specific type, specify a type when subclassing. For example:

    .. code-block:: python

        class MyRegistry(Registry[MyItemType]):
            ...

    Version Changed:
        3.1:
        Added support for specifying a registry item type when subclassing
        this registry.
    """

    #: The name of the items being registered.
    #:
    #: Type:
    #:     str
    item_name: Optional[str] = None

    #: A list of attributes that items can be looked up by.
    #:
    #: Type:
    #:     list of str
    lookup_attrs: Sequence[str] = []

    #: Error formatting strings for exceptions.
    #:
    #: Entries here override the global :py:data:`DEFAULT_ERRORS` dictionary
    #: for error messages.
    #:
    #: Type:
    #:     dict
    errors: Dict[str, str] = {}

    #: The default error formatting strings.
    #:
    #: If subclasses need to provide additional errors that can be overridden,
    #: they should copy :py:data:`DEFAULT_ERRORS` and set their copy on the
    #: subclass as this attribute.
    #:
    #: Type:
    #:     dict
    default_errors: Dict[str, str] = DEFAULT_ERRORS

    #: The error class indicating an already registered item.
    #:
    #: This must be a subclass of
    #: :py:class:`~djblets.registries.errors.AlreadyRegisteredError`.
    #:
    #: Type:
    #:     type
    already_registered_error_class: Type[AlreadyRegisteredError] = \
        AlreadyRegisteredError

    #: The lookup error exception class.
    #:
    #: This must be a subclass of
    #: :py:class:`~djblets.registries.errors.ItemLookupError`.
    #:
    #: Type:
    #:     type
    lookup_error_class: Type[ItemLookupError] = ItemLookupError

    def __init__(self) -> None:
        """Initialize the registry."""
        self._registry: Dict[str, Dict[object, RegistryItemType]] = {
            _attr_name: {}
            for _attr_name in self.lookup_attrs
        }
        self._populated: bool = False
        self._items: Set[RegistryItemType] = set()

    @property
    def populated(self) -> bool:
        """Whether or not the registry is populated.

        Returns:
            bool:
            Whether or not the registry is populated.
        """
        return self._populated

    def format_error(
        self,
        error_name: str,
        **error_kwargs,
    ) -> str:
        """Format an error message.

        Args:
            error_name (str):
                A symbolic name for the error, such as
                :py:data:`ALREADY_REGISTERED`.

            **error_kwargs (dict):
                The keyword arguments to provide to the error-specific
                formatting string.

        Returns:
            str:
            The formatted error message.

        Raises:
            ValueError:
                A registered error message for ``error_name`` could not be
                found.
        """
        fmt = self.errors.get(error_name, self.default_errors.get(error_name))

        if fmt is None:
            raise ValueError('%s.format_error: Unknown error: "%s".',
                             type(self), error_name)

        return fmt % error_kwargs

    def get(
        self,
        attr_name: str,
        attr_value: object,
    ) -> RegistryItemType:
        """Return an item by its attribute value.

        Args:
            attr_name (str):
                The attribute name to look up an item by.

            attr_value (object):
                The corresponding attribute value.

        Returns:
            object:
            The registered item.

        Raises:
            djblets.registries.errors.ItemLookupError:
                When a lookup is attempted with an unsupported attribute, or
                the item cannot be found, this exception is raised.
        """
        self.populate()

        try:
            attr_map = self._registry[attr_name]
        except KeyError:
            raise self.lookup_error_class(self.format_error(
                INVALID_ATTRIBUTE, attr_name=attr_name))

        try:
            return attr_map[attr_value]
        except KeyError:
            raise self.lookup_error_class(self.format_error(
                NOT_REGISTERED, attr_name=attr_name, attr_value=attr_value))

    def get_or_none(
        self,
        attr_name: str,
        attr_value: object,
    ) -> Optional[RegistryItemType]:
        """Return the requested registered item, or None if not found.

        Version Added:
            3.1

        Args:
            attr_name (str):
                The attribute name.

            attr_value (object):
                The attribute value.

        Returns:
            object:
            The matching registered item, if found. Otherwise, ``None`` is
            returned.
        """
        try:
            return self.get(attr_name, attr_value)
        except ItemLookupError:
            return None

    def register(
        self,
        item: RegistryItemType,
    ) -> None:
        """Register an item.

        Args:
            item (object):
                The item to register with the class.

        Raises:
            djblets.registries.errors.RegistrationError:
                Raised if the item is missing one of the required attributes.

            djblets.registries.errors.AlreadyRegisteredError:
                Raised if the item is already registered or if the item shares
                an attribute name, attribute value pair with another item in
                the registry.
        """
        self.populate()
        attr_values: Dict[str, object] = {}

        if item in self._items:
            raise self.already_registered_error_class(self.format_error(
                ALREADY_REGISTERED,
                item=item))

        registry_map = self._registry

        for attr_name in self.lookup_attrs:
            attr_map = registry_map[attr_name]

            try:
                attr_value = getattr(item, attr_name)

                if attr_value in attr_map:
                    raise self.already_registered_error_class(
                        self.format_error(ATTRIBUTE_REGISTERED,
                                          item=item,
                                          duplicate=attr_map[attr_value],
                                          attr_name=attr_name,
                                          attr_value=attr_value))

                attr_values[attr_name] = attr_value
            except AttributeError:
                raise RegistrationError(self.format_error(
                    MISSING_ATTRIBUTE,
                    item=item,
                    attr_name=attr_name))

        for attr_name, attr_value in attr_values.items():
            registry_map[attr_name][attr_value] = item

        self._items.add(item)

    def unregister_by_attr(
        self,
        attr_name: str,
        attr_value: object,
    ) -> None:
        """Unregister an item from the registry by an attribute.

        Args:
            attr_name (str):
                The name of the attribute.

            attr_value (object):
                The attribute value.

        Raises:
            djblets.registries.errors.ItemLookupError:
                Raised if the attribute value is not found in the registry.
        """
        self.populate()

        try:
            attr_map = self._registry[attr_name]
        except KeyError:
            raise self.lookup_error_class(self.format_error(
                INVALID_ATTRIBUTE, attr_name=attr_name))
        try:
            item = attr_map[attr_value]
        except KeyError:
            raise self.lookup_error_class(self.format_error(
                NOT_REGISTERED, attr_name=attr_name, attr_value=attr_value))

        self.unregister(item)

    def unregister(
        self,
        item: RegistryItemType,
    ) -> None:
        """Unregister an item from the registry.

        Args:
            item (object):
                The item to unregister. This must be present in the registry.

        Raises:
            djblets.registries.errors.ItemLookupError:
                Raised if the item is not found in the registry.
        """
        self.populate()

        try:
            self._items.remove(item)
        except KeyError:
            raise self.lookup_error_class(self.format_error(UNREGISTER,
                                                            item=item))

        registry_map = self._registry

        for attr_name in self.lookup_attrs:
            attr_value = getattr(item, attr_name)
            del registry_map[attr_name][attr_value]

    def populate(self) -> None:
        """Ensure the registry is populated.

        Calling this method when the registry is populated will have no effect.
        """
        if self._populated:
            return

        self._populated = True

        for item in self.get_defaults():
            self.register(item)

        registry_populating.send(sender=type(self),
                                 registry=self)

    def get_defaults(self) -> Iterable[RegistryItemType]:
        """Return the default items for the registry.

        This method should be overridden by a subclass.

        Returns:
            list:
            The default items for the registry.
        """
        return []

    def reset(self) -> None:
        """Unregister all items and mark the registry unpopulated.

        This will result in the registry containing no entries. Any call to a
        method that would populate the registry will repopulate it.
        """
        if self._populated:
            for item in self._items.copy():
                self.unregister(item)

            self._populated = False

        assert len(self._items) == 0

    def __iter__(self) -> Iterator[RegistryItemType]:
        """Iterate through all items in the registry.

        This method does not provide a stable ordering.

        Yields:
            object:
            The items registered in this registry.
        """
        self.populate()

        for item in self._items:
            yield item

    def __len__(self) -> int:
        """Return the number of items in the registry.

        Returns:
            int:
            The number of items in the registry.
        """
        self.populate()
        return len(self._items)

    def __contains__(
        self,
        item: RegistryItemType,
    ) -> bool:
        """Return whether or not the item is contained in the registry.

        Args:
            item (object):
                The item to look for.

        Returns:
            bool:
            Whether or not the item is contained in the registry.
        """
        self.populate()
        return item in self._items


class EntryPointRegistry(Registry[RegistryItemType]):
    """A registry that auto-populates from an entry-point."""

    #: The entry point name.
    #:
    #: Type:
    #:     str
    entry_point: Optional[str] = None

    def get_defaults(self) -> Iterable[RegistryItemType]:
        """Yield the values from the entry point.

        Yields:
            object:
            The object from the entry point.
        """
        if self.entry_point is not None:
            entry_points = iter_entry_points(self.entry_point)

            for ep in entry_points:
                try:
                    yield self.process_value_from_entry_point(ep)
                except Exception as e:
                    logging.exception(self.format_error(LOAD_ENTRY_POINT,
                                                        entry_point=ep.name,
                                                        error=e))

    def process_value_from_entry_point(
        self,
        entry_point: EntryPoint,
    ) -> RegistryItemType:
        """Return the item to register from the entry point.

        By default, this returns the loaded entry point.

        Args:
            entry_point (pkg_resources.EntryPoint):
                The entry point.

        Returns:
            object:
            The processed entry point value.
        """
        return entry_point.load()


class OrderedRegistry(Registry[RegistryItemType]):
    """A registry that keeps track of registration order."""

    def __init__(self) -> None:
        """Initialize the OrderedRegistry"""
        super(OrderedRegistry, self).__init__()

        self._by_id: Dict[int, RegistryItemType] = {}
        self._key_order: List[int] = []

    def register(
        self,
        item: RegistryItemType,
    ) -> None:
        """Register an item.

        Args:
            item (object):
                The item to register with the class.

        Raises:
            djblets.registries.errors.RegistrationError:
                Raised if the item is missing one of the required attributes.

            djblets.registries.errors.AlreadyRegisteredError:
                Raised if the item is already registered or if the item shares
                an attribute name, attribute value pair with another item in
                the registry.
        """
        super(OrderedRegistry, self).register(item)
        item_id = id(item)
        self._key_order.append(item_id)
        self._by_id[item_id] = item

    def unregister(
        self,
        item: RegistryItemType,
    ) -> None:
        """Unregister an item from the registry.

        Args:
            item (object):
                The item to unregister. This must be present in the registry.

        Raises:
            djblets.registries.errors.ItemLookupError:
                Raised if the item is not found in the registry.
        """
        super(OrderedRegistry, self).unregister(item)
        item_id = id(item)
        del self._by_id[item_id]
        self._key_order.remove(item_id)

    def __iter__(self) -> Iterator[RegistryItemType]:
        """Yield the items in the order they were registered.

        Yields:
            object:
            The registered items.
        """
        self.populate()

        by_id = self._by_id

        for key in self._key_order:
            yield by_id[key]

    def __getitem__(
        self,
        index: int,
    ) -> RegistryItemType:
        """Return an item by its registered index.

        Args:
            index (int):
                The position at which the item was registered. This is 0-based
                and negative indices are supported.

        Returns:
            object:
            The requested item.

        Raises:
            IndexError:
                This exception is raised if the requested index is out of
                range.

            TypeError:
                This exception is raised if the requested index is not an
                integer.
        """
        if not isinstance(index, int):
            raise TypeError('Index is not an integer (is %s).'
                            % type(index).__name__)

        # We don't have to call populate() because calling len() will.
        length = len(self)

        if index < 0:
            index += length

        if index > length:
            raise IndexError('Index is out of range.')

        return self._by_id[self._key_order[index]]



"""Utilities related to django.contrib.admin fieldsets."""


def filter_fieldsets(form, admin=None, fieldsets=None, exclude=None,
                     exclude_collapsed=True):
    """Filter fieldsets.

    This method allows us to filter fieldsets from a ModelAdmin to exclude
    fields (or an entire fieldset).

    Args:
        form (type or django.forms.Form):
            The form (or form class) to retrieve fieldsets for.

        admin (django.contrib.admin.ModelAdmin, optional):
            The model admin to retrieve fieldsets from. If this argument is not
            provided, ``fieldsets`` will be used instead.

        fieldsets (tuple, optional):
            The fieldsets to use.

        exclude (list of unicode, optional):
            An optional list of fields to exclude.

        exclude_collapsed (bool, optional):
            Whether or not to exclude fieldsets marked as collapsed by default.

    Yields:
        dict:
        An entry for each field set that should be rendered.

        If a field set would have no rendered fields, it is not yielded.
    """
    if admin is None and fieldsets is None:
        raise ValueError(
            'filter_fieldsets: either admin or fieldsets must be provided.'
        )

    if exclude is None:
        exclude = []

    exclude.extend(list(form._meta.exclude or []))

    if fieldsets is None:
        assert hasattr(admin, 'fieldsets'), 'admin.fieldsets is undefined.'
        assert iter(admin.fieldsets), 'admin.fieldsets should be iterable.'
        fieldsets = admin.fieldsets

    for name, data in fieldsets:
        fieldset = data.copy()

        if exclude_collapsed and 'collapse' in data.get('classes', ()):
            continue

        fieldset['fields'] = [
            field_name
            for field_name in data['fields']
            if field_name not in exclude
        ]

        if fieldset['fields']:
            yield (name, fieldset)



"""Unit tests for djblets.util.decorators."""

from djblets.testing.testcases import TestCase
from djblets.util.decorators import cached_property, optional_decorator


class DecoratorTests(TestCase):
    """Unit tests for djblets.util.decorators."""

    def test_cached_property(self):
        """Testing @cached_property retains attributes and docstring"""
        class MyClass(object):
            def expensive_method(self, state=[0]):
                state[0] += 1

                return state[0]

            def my_prop1(self):
                """This is my docstring."""
                return self.expensive_method()

            my_prop1.some_attr = 105
            my_prop1 = cached_property(my_prop1)

            @cached_property
            def my_prop2(self):
                """Another one!"""
                return 'foo'

        instance = MyClass()

        self.assertEqual(instance.my_prop1, 1)
        self.assertEqual(instance.my_prop1, 1)
        self.assertEqual(instance.my_prop2, 'foo')

        prop1_instance = instance.__class__.__dict__['my_prop1']
        self.assertEqual(prop1_instance.__name__, 'my_prop1')
        self.assertEqual(prop1_instance.__doc__, 'This is my docstring.')
        self.assertEqual(getattr(prop1_instance, 'some_attr'), 105)

        prop2_instance = instance.__class__.__dict__['my_prop2']
        self.assertEqual(prop2_instance.__name__, 'my_prop2')
        self.assertEqual(prop2_instance.__doc__, 'Another one!')
        self.assertFalse(hasattr(prop2_instance, 'some_attr'))

    def test_optional_decorator(self):
        """Testing @optional_decorator"""
        def predicate(x):
            return x != 42

        def decorator(f):
            def decorated(x):
                return f(x) + 1

            return decorated

        @optional_decorator(decorator, predicate)
        def f(x):
            return x

        self.assertEqual(f(0), 1)
        self.assertEqual(f(1), 2)
        self.assertEqual(f(42), 42)



"""Unit tests for djblets.extensions.hooks.BaseRegistryHook."""

from djblets.extensions.extension import Extension
from djblets.extensions.hooks import ExtensionHookPoint, BaseRegistryHook
from djblets.extensions.testing import ExtensionTestCaseMixin
from djblets.registries.registry import Registry
from djblets.testing.testcases import TestCase


class MyTestExtension(Extension):
    pass


class BaseRegistryHookTests(ExtensionTestCaseMixin, TestCase):
    """Unit tests for djblets.extensions.hooks.BaseRegistryHook."""

    extension_class = MyTestExtension

    class DummyRegistry(Registry):
        lookup_attrs = ('foo_id',)

    class DummyItem(object):
        def __init__(self, foo_id):
            self.foo_id = foo_id

    def setUp(self):
        super(BaseRegistryHookTests, self).setUp()

        self.registry = self.DummyRegistry()

        class DummyRegistryHook(BaseRegistryHook,
                                metaclass=ExtensionHookPoint):
            registry = self.registry

        self.hook_cls = DummyRegistryHook

    def test_hook_register(self):
        """Testing BaseRegistryHook item registration"""
        self.assertEqual(list(self.registry), [])

        item = self.DummyItem(123)
        self.hook_cls(self.extension, item)

        self.assertIn(item, self.registry)

    def test_hook_unregister(self):
        """Testing BaseRegistryHook item unregistration"""
        self.assertEqual(list(self.registry), [])
        item = self.DummyItem(123)
        self.hook_cls(self.extension, item)

        self.extension.shutdown()
        self.assertEqual(list(self.registry), [])



"""Authentication support for using API tokens.

This provides two classes needed for integrating API token support.

:py:class:`TokenAuthBackendMixin` must be subclassed or mixed into an
authentication backend and registered in
:django:setting:`AUTHENTICATION_BACKENDS`.

:py:class:`WebAPITokenAuthBackend` must then be added to the API authentication
list at ``settings.WEB_API_AUTH_BACKENDS``.
"""

import logging

from django.contrib.auth import get_backends
from django.utils import timezone
from django.utils.formats import localize
from django.utils.translation import gettext as _

from djblets.webapi.auth import WebAPIAuthBackend
from djblets.webapi.errors import LOGIN_FAILED
from djblets.webapi.signals import webapi_token_expired


logger = logging.getLogger(__name__)


class TokenAuthBackendMixin(object):
    """Mixin for a standard auth backend for API token authentication.

    This will handle authenticating users and their API tokens for API
    requests. It's only used for API requests that specify a username and a
    token.

    This class is meant to be subclassed and mixed in to another auth backend.
    Subclasses must define :py:attr:`api_token_model`.
    """

    #: The API token model to use for any token lookups.
    api_token_model = None

    def authenticate(self, request, token=None, **kwargs):
        """Authenticate a user, given a token ID.

        Args:
            request (django.http.HttpRequest, unused):
                The request object.

            token (unicode, optional):
                The API token ID to authenticate with.

            **kwargs (dict, unused):
                Keyword arguments for future expansion.

        Returns:
            User:
            The resulting user, if a token matched, or ``None`` otherwise.
        """
        if not token:
            return None

        # Find the WebAPIToken matching the token parameter passed in.
        # Once we have it, we'll need to perform some additional checks on
        # the user.
        q = self.api_token_model.objects.filter(token=token)
        q = q.select_related('user')

        try:
            webapi_token = q.get()
        except self.api_token_model.DoesNotExist:
            return None

        user = webapi_token.user

        if not user.is_active:
            return None

        # Store this temporarily. We'll be using it to store some session
        # state.
        user._webapi_token = webapi_token

        return user

    def validate_token(self, request, token, **kwargs):
        """Check that the token is valid to use for authentication.

        This will check if the token is invalid or expired. If so it will
        return a dictionary containing an error message with details on whether
        the token is invalid or expired and any appropriate HTTP headers to
        send to the client. If the token is valid to use for authentication
        this will return ``None``.

        Version Added:
            3.0

        Args:
            request (django.http.HttpRequest):
                The request object.

            token (str):
                The API token ID to validate.

            **kwargs (dict, unused):
                Additional keyword arguments.

        Returns:
            dict or None:
            A dictionary containing the following keys:

            ``error_message`` (str):
                An error message explaining why the token cannot be used
                for authentication.

            ``headers`` (dict):
                A dictionary of HTTP headers to send to the client.

            These are meant to be used as the ``error_message`` and ``header``
            values in the return type of :py:meth:`djblets.webapi.auth.
            backends.base.WebAPIAuthBackend.authenticate`.

            If the token is valid to use for authentication this will return
            ``None``.
        """
        q = self.api_token_model.objects.filter(token=token)
        log_extra = {
            'request': request,
        }

        try:
            webapi_token = q.get()
        except self.api_token_model.DoesNotExist:
            logger.debug('API Login failed. Token not found.',
                         extra=log_extra)

            return {
                'error_message': None,
                'headers': None,
            }

        if not webapi_token.valid:
            logger.debug('API Login failed for %s. The token is invalid.',
                         webapi_token.user.username,
                         extra=log_extra)

            if webapi_token.invalid_reason:
                error_message = (
                    _('This API token became invalid on %s: %s')
                    % (localize(webapi_token.invalid_date),
                       webapi_token.invalid_reason)
                )
            else:
                error_message = (
                    _('This API token became invalid on %s.')
                    % localize(webapi_token.invalid_date)
                )

            return {
                'error_message': error_message,
                'headers': LOGIN_FAILED.headers(request).copy(),
            }
        elif webapi_token.is_expired():
            logger.debug('API Login failed for %s. The token is expired.',
                         webapi_token.user.username,
                         extra=log_extra)

            webapi_token_expired.send(instance=webapi_token,
                                      sender=type(webapi_token))

            return {
                'error_message': (
                    _('This API token expired on %s.')
                    % localize(webapi_token.expires)
                ),
                'headers': LOGIN_FAILED.headers(request).copy(),
            }

        return None


class WebAPITokenAuthBackend(WebAPIAuthBackend):
    """Authenticates users using their generated API token.

    This will check the ``HTTP_AUTHORIZATION`` header for a ``token <token>``
    value. If found, it will attempt to find the user that owns the
    token, and authenticate that user.
    """

    def get_credentials(self, request):
        """Return credentials for the token.

        If the request is attempting to authenticate with a token, this
        will return a dictionary containing the token in a ``token`` key.

        Args:
            request (HttpRequest): The HTTP request from the client.

        Returns:
            dict:
            A dictionary containing the token in a ``token`` key, if the
            client requested a token for authentication.
        """
        http_auth = request.META['HTTP_AUTHORIZATION']
        parts = http_auth.split(' ')

        if parts[0] != 'token':
            return None

        if len(parts) != 2:
            logger.warning('WebAPITokenAuthBackend: Missing token in '
                           'HTTP_AUTHORIZATION header %s',
                           http_auth, extra={'request': request})
            return None

        return {
            'token': parts[1],
        }

    def login_with_credentials(self, request, **credentials):
        """Log the user in with the given credentials.

        This performs the standard authentication operations, and then
        stores some session state for any restrictions specified by the
        token.

        Args:
            request (HttpRequest):
                The HTTP request from the client.

            credentials (dict):
                The credentials data from the request.

        Returns:
            tuple or None:
            See the return type in :py:meth:`djblets.webapi.auth.backends.
            base.WebAPIAuthBackend.authenticate`.
        """
        result = super().login_with_credentials(request, **credentials)

        if result[0]:
            user = request.user
            webapi_token = user._webapi_token
            del user._webapi_token

            webapi_token.last_used = timezone.now()
            webapi_token.save_base(update_fields=('last_used',))

            request.session['webapi_token_id'] = webapi_token.pk
            request._webapi_token = webapi_token

            if webapi_token.is_deprecated():
                deprecation_header = {
                    'X-API-Token-Deprecated': (
                        'This token uses a deprecated format. The token can '
                        'still be used, but you should remove it and '
                        'generate a new one to take advantage of security '
                        'improvements.'
                    )
                }

                # Keep any headers that were already set.
                headers = result[2] or {}
                headers.update(deprecation_header)

                result = (result[0], result[1], headers)

        return result

    def validate_credentials(self, request, **credentials):
        """Validate that credentials are valid.

        This will run through authentication backends to check whether
        the API token is valid to use. If the token is not valid to use this
        will stop the authentication process and return a message and headers
        on why it cannot be used.

        Version Added:
            3.0

        Args:
            request (django.http.HttpRequest):
                The HTTP request from the client.

            credentials (dict):
                All credentials provided by :py:meth:`get_credentials`.

        Returns:
            tuple or None:
            See the return type in :py:meth:`WebAPIAuthBackend.authenticate`.
        """
        result = super().validate_credentials(request, **credentials)

        if result is not None:
            return result

        for backend in get_backends():
            try:
                result = backend.validate_token(request,
                                                credentials.get('token'))

                if result is not None:
                    return (False,
                            result.get('error_message'),
                            result.get('headers'))
            except AttributeError:
                continue

        return None



from itertools import chain

from django.core.management.base import BaseCommand, CommandError
from django.utils.translation import gettext as _

from djblets.extensions.errors import (InstallExtensionError,
                                       InvalidExtensionError)
from djblets.extensions.manager import get_extension_managers


class Command(BaseCommand):
    """Install the extension media.

    This command will install the extension media for either a specific
    extension (specified with the ``--extension-id`` flag) or all installed
    extensions. Installation of the media can be forced (i.e., no version
    checking will be done.
    """

    def add_arguments(self, parser):
        """Add arguments to the command.

        Args:
            parser (object):
                The argument parser to add to.
        """
        parser.add_argument(
            '--extension-id',
            dest='extension_id',
            default=None,
            help=_('An optional extension id'))

        parser.add_argument(
            '--force',
            dest='force',
            action='store_true',
            default=False,
            help=_('Force installation of extension media'))

    def handle(self, *args, **options):
        managers = get_extension_managers()

        force_install = options['force']

        if options['extension_id']:
            extensions = [self._find_extension(options['extension_id'],
                                               managers)]
        else:
            extensions = chain(
                (extension, manager)
                for manager in managers
                for extension in manager.get_enabled_extensions()
            )

        for extension, manager in extensions:
            try:
                manager.install_extension_media(extension, force_install)
            except InstallExtensionError as e:
                raise CommandError('Could not install extension media: %s'
                                   % e)

    def _find_extension(self, extension_id, managers):
        """Find an extension with the given ID in the managers.

        Args:
            extension_id (unicode):
                The extension's ID.

            managers (list):
                A list of :py:class:`~django.extensions.extension.Extension`
                classes (not instances).

        Returns:
            type:
            The specific :py:class:`~django.extensions.extension.Extension`
            class.
        """
        for manager in managers:
            try:
                return manager.get_enabled_extension(extension_id), manager
            except InvalidExtensionError:
                pass

        raise CommandError('No such extension: %s' % extension_id)



"""Template loaders for extensions."""

from django.template import Origin, TemplateDoesNotExist
from django.template.loaders.base import Loader as BaseLoader
from pkg_resources import _manager as manager

from djblets.extensions.manager import get_extension_managers


class ExtensionOrigin(Origin):
    """An origin for a template in an extension.

    Version Added:
        3.0
    """

    def __init__(self, package, resource, *args, **kwargs):
        """Initialize the origin.

        Args:
            package (unicode):
                The name of the package providing the template.

            resource (unicode):
                The resource path within the package.

            *args (tuple):
                Positional arguments to pass to the parent.

            **kwargs (dict):
                Keyword arguments to pass to the parent.
        """
        self.package = package
        self.resource = resource

        super(ExtensionOrigin, self).__init__(*args, **kwargs)


class Loader(BaseLoader):
    """Loads templates found within an extension.

    This will look through all enabled extensions and attempt to fetch
    the named template under the :file:`templates` directory within the
    extension's package.

    This should be added last to the list of template loaders.

    .. versionadded:: 0.9
    """

    is_usable = manager is not None

    def get_contents(self, origin):
        """Return the contents of a template.

        Args:
            origin (ExtensionOrigin):
                The origin of the template.

        Returns:
            unicode:
            The resulting template contents.

        Raises:
            TemplateDoesNotExist:
                The template could not be found.
        """
        try:
            data = manager.resource_string(origin.package, origin.resource)
            return data.decode('utf-8')
        except Exception:
            raise TemplateDoesNotExist(origin)

    def get_template_sources(self, template_name):
        """Load templates from enabled extensions.

        Args:
            template_name (unicode):
                The name of the template to load.

        Yields:
            ExtensionOrigin:
            Each possible location for the template.
        """
        if manager:
            resource = "templates/" + template_name

            for extmgr in get_extension_managers():
                for ext in extmgr.get_enabled_extensions():
                    package = ext.info.app_name

                    yield ExtensionOrigin(
                        package=package,
                        resource=resource,
                        name='extension:%s:%s' % (package, resource),
                        template_name=template_name,
                        loader=self)
